'''
Strain class
'''
from logging import getLogger
import sys

import numpy as np
from pymatgen.core import Structure
from pymatgen.core.periodic_table import DummySpecie

from ...IO import read_input as rin
from ...util.struc_util import sort_by_atype, sort_by_atype_mol, check_distance, cal_g, find_site


logger = getLogger('cryspy')

class Strain:
    '''
    strain

    # ---------- instance methods
    gen_child(self, struc)
    '''

    def __init__(self, mindist):
        self.mindist = mindist

    def gen_child(self, struc):
        '''
        generate child struture

        # ---------- return
        (if success) self.child:
        (if fail) None:
        '''
        # ---------- init
        cnt = 0
        lat_mat = struc.lattice.matrix.T    # lattice vector as matrix
        # ---------- generate strained structure
        while True:
            # ------ strain matrix
            strain_matrix = np.empty([3, 3])
            for i in range(3):
                for j in range(3):
                    if i <= j:
                        if i == j:
                            strain_matrix[i][j] = 1.0 + np.random.normal(
                                loc=0.0, scale=rin.sigma_st)
                        else:
                            strain_matrix[i][j] = np.random.normal(
                                loc=0.0, scale=rin.sigma_st)/2.0
                            strain_matrix[j][i] = strain_matrix[i][j]
            # ------ strained lattice
            strained_lattice = np.dot(strain_matrix, lat_mat).T
            # ------ child
            self.child = Structure(strained_lattice, struc.species,
                                   struc.frac_coords)
            # ------ scale lattice
            self.child.scale_lattice(struc.volume)
            # ------ check distance
            success, mindist_ij, dist = check_distance(self.child,
                                                       rin.atype,
                                                       self.mindist)
            if success:
                self.child = sort_by_atype(self.child, rin.atype)
                return self.child
            else:
                type0 = rin.atype[mindist_ij[0]]
                type1 = rin.atype[mindist_ij[1]]
                logger.warning(f'mindist in strain: {type0} - {type1}, {dist}. retry.')
                cnt += 1
                if cnt >= rin.maxcnt_ea:
                    self.child = None
                    return None    # change parent

    def gen_child_mol(self, struc, mol_id):
        '''
        generate child structures for mol

        # ---------- return
        (if success) self.child:
        (if fail) None:
        '''
        # ---------- init
        cnt = 0
        lat_mat = struc.lattice.matrix.T    # lattice vector as matrix
        # ---------- generate strained structure
        while True:
            # ------ strain matrix
            strain_matrix = np.empty([3, 3])
            for i in range(3):
                for j in range(3):
                    if i <= j:
                        if i == j:
                            strain_matrix[i][j] = 1.0 + np.random.normal(
                                loc=0.0, scale=rin.sigma_st)
                        else:
                            strain_matrix[i][j] = np.random.normal(
                                loc=0.0, scale=rin.sigma_st)/2.0
                            strain_matrix[j][i] = strain_matrix[i][j]
            # ------ strained lattice
            strained_lattice = np.dot(strain_matrix, lat_mat).T
            # -- generate dummy structure
            dum_coords = cal_g(struc, mol_id[0], mol_id[1], mol_id[2])
            dum_species = []
            for i in range(len(dum_coords)):
                dum_species.append(DummySpecie("X{}".format(i)))
            dum_struc = Structure(strained_lattice, dum_species, dum_coords)
            # ------ scale lattice
            dum_struc.scale_lattice(struc.volume)
            # --calcurate atom coords to replace molecure
            fix_frac_coords, fix_group_id, fix_species, fix_mol_id = find_site(struc,
                                                                               mol_id[0],
                                                                               mol_id[1],
                                                                               mol_id[2])
            rm_spe = []
            strained_mol_id = []
            strained_group_id = []
            append_spe_co = []
            append_mol_gr = []
            for i, dum in enumerate(dum_struc):
                for j, mid in enumerate(fix_group_id):
                    if mid == i:
                        coord = (struc.lattice.get_cartesian_coords(fix_frac_coords[j])
                                 - struc.lattice.get_cartesian_coords(dum_coords[i])
                                 + dum_struc.cart_coords[i])
                        append_spe_co.append([fix_species[j], coord])
                        append_mol_gr.append([fix_mol_id[j], fix_group_id[j]])
                rm_spe.append(dum.specie)
            # replace mol
            for spe_co, mol_gr in zip(append_spe_co, append_mol_gr):
                dum_struc.append(spe_co[0], spe_co[1], coords_are_cartesian=True)
                strained_mol_id.append(mol_gr[0])
                strained_group_id.append(mol_gr[1])
            # remove dummy
            dum_struc.remove_species(rm_spe)
            # ------ child
            self.child = dum_struc
            # ------ check distance
            success, mindist_ij, dist = check_distance(self.child,
                                                       rin.atype,
                                                       self.mindist)
            if success:
                self.child, self.mol_id, self.group_id = sort_by_atype_mol(self.child, rin.atype,
                                                                           strained_mol_id, strained_group_id)
                return self.child, [self.mol_id, self.group_id, mol_id[2]]
            else:
                type0 = rin.atype[mindist_ij[0]]
                type1 = rin.atype[mindist_ij[1]]
                logger.warning(f'mindist in permutation: {type0} - {type1}, {dist}. retry.')
                cnt += 1
                if cnt >= rin.maxcnt_ea:
                    self.child = None
                    return None    # change parent



'''
Save data in ./data/pkl_data/xxx
'''

import pickle


def load_init_struc():
    with open('./data/pkl_data/init_struc_data.pkl', 'rb') as f:
        init_struc_data = pickle.load(f)
    return init_struc_data


def save_init_struc(init_struc_data):
    with open('./data/pkl_data/init_struc_data.pkl', 'wb') as f:
        pickle.dump(init_struc_data, f)


def load_opt_struc():
    with open('./data/pkl_data/opt_struc_data.pkl', 'rb') as f:
        opt_struc_data = pickle.load(f)
    return opt_struc_data


def save_opt_struc(opt_struc_data):
    with open('./data/pkl_data/opt_struc_data.pkl', 'wb') as f:
        pickle.dump(opt_struc_data, f)


def load_rslt():
    with open('./data/pkl_data/rslt_data.pkl', 'rb') as f:
        rslt_data = pickle.load(f)
    return rslt_data


def save_rslt(rslt_data):
    with open('./data/pkl_data/rslt_data.pkl', 'wb') as f:
        pickle.dump(rslt_data, f)


def load_kpt():
    with open('./data/pkl_data/kpt_data.pkl', 'rb') as f:
        kpt_data = pickle.load(f)
    return kpt_data


def save_kpt(kpt_data):
    with open('./data/pkl_data/kpt_data.pkl', 'wb') as f:
        pickle.dump(kpt_data, f)


def load_energy_step():
    with open('./data/pkl_data/energy_step_data.pkl', 'rb') as f:
        energy_step_data = pickle.load(f)
    return energy_step_data


def save_energy_step(energy_step_data):
    with open('./data/pkl_data/energy_step_data.pkl', 'wb') as f:
        pickle.dump(energy_step_data, f)


def load_struc_step():
    with open('./data/pkl_data/struc_step_data.pkl', 'rb') as f:
        struc_step_data = pickle.load(f)
    return struc_step_data


def save_struc_step(struc_step_data):
    with open('./data/pkl_data/struc_step_data.pkl', 'wb') as f:
        pickle.dump(struc_step_data, f)


def load_force_step():
    with open('./data/pkl_data/force_step_data.pkl', 'rb') as f:
        force_step_data = pickle.load(f)
    return force_step_data


def save_force_step(force_step_data):
    with open('./data/pkl_data/force_step_data.pkl', 'wb') as f:
        pickle.dump(force_step_data, f)


def load_stress_step():
    with open('./data/pkl_data/stress_step_data.pkl', 'rb') as f:
        stress_step_data = pickle.load(f)
    return stress_step_data


def save_stress_step(stress_step_data):
    with open('./data/pkl_data/stress_step_data.pkl', 'wb') as f:
        pickle.dump(stress_step_data, f)


def load_rs_id():
    with open('./data/pkl_data/RS_id_data.pkl', 'rb') as f:
        rs_id_data = pickle.load(f)
    return rs_id_data


def save_rs_id(rs_id_data):
    with open('./data/pkl_data/RS_id_data.pkl', 'wb') as f:
        pickle.dump(rs_id_data, f)


def load_bo_id():
    with open('./data/pkl_data/BO_id_data.pkl', 'rb') as f:
        bo_id_data = pickle.load(f)
    return bo_id_data


def save_bo_id(bo_id_data):
    with open('./data/pkl_data/BO_id_data.pkl', 'wb') as f:
        pickle.dump(bo_id_data, f)


def load_bo_data():
    with open('./data/pkl_data/BO_data.pkl', 'rb') as f:
        bo_data = pickle.load(f)
    return bo_data


def save_bo_data(bo_data):
    with open('./data/pkl_data/BO_data.pkl', 'wb') as f:
        pickle.dump(bo_data, f)


def load_laqa_id():
    with open('./data/pkl_data/LAQA_id_data.pkl', 'rb') as f:
        laqa_id_data = pickle.load(f)
    return laqa_id_data


def save_laqa_id(laqa_id_data):
    with open('./data/pkl_data/LAQA_id_data.pkl', 'wb') as f:
        pickle.dump(laqa_id_data, f)


def load_laqa_data():
    with open('./data/pkl_data/LAQA_data.pkl', 'rb') as f:
        laqa_data = pickle.load(f)
    return laqa_data


def save_laqa_data(laqa_data):
    with open('./data/pkl_data/LAQA_data.pkl', 'wb') as f:
        pickle.dump(laqa_data, f)


def load_ea_id():
    with open('./data/pkl_data/EA_id_data.pkl', 'rb') as f:
        ea_id_data = pickle.load(f)
    return ea_id_data


def load_struc_mol_id():
    with open('./data/pkl_data/struc_mol_id_data.pkl', 'rb') as f:
        struc_mol_id = pickle.load(f)
    return struc_mol_id


def save_struc_mol_id(struc_mol_id):
    with open('./data/pkl_data/struc_mol_id_data.pkl', 'wb') as f:
        pickle.dump(struc_mol_id, f)


def save_ea_id(ea_id_data):
    with open('./data/pkl_data/EA_id_data.pkl', 'wb') as f:
        pickle.dump(ea_id_data, f)


def load_ea_data():
    with open('./data/pkl_data/EA_data.pkl', 'rb') as f:
        ea_data = pickle.load(f)
    return ea_data


def save_ea_data(ea_data):
    with open('./data/pkl_data/EA_data.pkl', 'wb') as f:
        pickle.dump(ea_data, f)


def load_ext_opt_struc():
    with open('./ext/calc_data/ext_opt_struc_data.pkl', 'rb') as f:
        ext_opt_struc_data = pickle.load(f)
    return ext_opt_struc_data


def load_ext_energy():
    with open('./ext/calc_data/ext_energy_data.pkl', 'rb') as f:
        ext_energy_data = pickle.load(f)
    return ext_energy_data



'''
Select parents in evolutionary algorithm
'''

from logging import getLogger

import numpy as np
from pymatgen.analysis.structure_matcher import StructureMatcher

from ...IO import read_input as rin


logger = getLogger('cryspy')

class Select_parents:
    '''
    select parents

    # ---------- args
    struc_data (dict or list): structure data
        You may include None in struc_data
        if type of struc_data is list,
            struc_data is converted into dict type
            as {0: struc_data_0, 1: struc_data_1, ...}

    fitness (dict, list, or np.ndarray): fitness
        You may include None and np.nan in fitness
        if type of fitness is list or np.ndarray,
            fitness is converted into dict type
            as {0: fitness_0, 1: fitness_1, ...}

    elite_struc (dict or None): structure data of elite

    elite_fitness (dict or None): fitness of elite

    n_fittest (int): number of data which can survive
        if 0 (deault), all candidates can survive
        n_fittest <-- rin.n_fittest or rin.n_elite

    # ---------- instance methods
    self.set_tournament(self)

    self.set_roulette(self)

    self.get_parents(n_parent)
    '''

    def __init__(self, struc_data, fitness,
                 elite_struc=None, elite_fitness=None, n_fittest=0):
        # ---------- check args
        # ------ data
        self.struc_data, self.fitness = self._check_data(struc_data,
                                                         fitness,
                                                         elite_struc,
                                                         elite_fitness)
        # ------ None and np.nan --> inf or -inf for fitness
        for key, value in self.fitness.items():
            if value is None or np.isnan(value):
                self.fitness[key] = -np.inf if rin.fit_reverse else np.inf
            # ---- emax_ea
            if rin.emax_ea is not None:
                if value > rin.emax_ea:
                    self.fitness[key] = -np.inf if rin.fit_reverse else np.inf
                    logger.info(f'Eliminate ID {key}: {value} > emax_ea')
            # ---- emin_ea
            if rin.emin_ea is not None:
                if value < rin.emin_ea:
                    self.fitness[key] = -np.inf if rin.fit_reverse else np.inf
                    logger.info(f'Eliminate ID {key}: {value} < emin_ea')
        # ------ self
        self.n_fittest = n_fittest
        # ---------- ranking of fitness: list of id
        self.ranking = sorted(self.fitness, key=self.fitness.get,
                              reverse=rin.fit_reverse)
        #print('ranking', self.ranking)
        # ---------- remove duplicated structures and
        #                cut by survival of the fittest
        self._dedupe()    # get self.ranking_dedupe

    def _check_data(self, struc_data, fitness, elite_struc, elite_fitness):
        # ---------- struc_data and fitness
        if isinstance(struc_data, dict) and isinstance(fitness, dict):
            pass    # if dict, allow len(struc_data) != len(fitness)
        elif isinstance(struc_data, dict) and (
                isinstance(fitness, list) or isinstance(fitness, np.ndarray)):
            logger.error('struc_data is dict,'
                            ' so fitness should also be dict')
        elif isinstance(fitness, dict) and (
                isinstance(struc_data, list)
                or isinstance(struc_data, np.ndarray)):
            logger.error('fitness is dict,'
                            ' so struc_data should also be dict')
        elif isinstance(struc_data, list) and (
                isinstance(fitness, list) or isinstance(fitness, np.ndarray)):
            # ------ check number of data
            if not len(struc_data) == len(fitness):
                logger.error('not len(struc_data) == len(fitness)')
            # ------ convert
            struc_data = {i: struc_data[i] for i in range(len(struc_data))}
            fitness = {i: fitness[i] for i in range(len(fitness))}
        else:
            logger.error('Type of struc_data and fitness is wrong')
        # ---------- elite
        if isinstance(elite_struc, dict) and isinstance(elite_fitness, dict):
            # ------ check number of data
            if not len(elite_struc) == len(elite_fitness):
                logger.error('not len(elite_struc) == len(elite_fitness)')
            if None in elite_struc.values():
                logger.error('elite_struc includes None')
            if (None in elite_fitness.values()) or (
                    np.nan in elite_fitness.values()):
                logger.error('elite_fitness includes None or np.nan')
            # ------ add elite to data
            struc_data.update(elite_struc)
            fitness.update(elite_fitness)
        elif elite_struc is elite_fitness is None:
            pass
        else:
            logger.error('elite_struc and elite_fitness'
                            ' must be dict or None')
        # ---------- return
        return struc_data, fitness

    def _dedupe(self):
        # ---------- remove duplicated structure data
        # ------ initialize
        ncheck = 5
        self.ranking_dedupe = []
        smatcher = StructureMatcher()    # instantiate
        # ------ register not dupulicated data
        for i_id in self.ranking:
            # -- init dupl_flag
            dupl_flag = False
            # -- for structure is None
            if self.struc_data[i_id] is None:
                continue    # next i_id
            # -- duplication check
            for j_id in self.ranking_dedupe[:-(ncheck+1):-1]:
                if smatcher.fit(self.struc_data[i_id], self.struc_data[j_id]):
                    dupl_flag = True
                    break
            # -- register or skip
            if dupl_flag:
                continue    # next i_id
            else:
                self.ranking_dedupe.append(i_id)
                # n_fittest
                if self.n_fittest > 0:
                    if len(self.ranking_dedupe) == self.n_fittest:
                        break
        # ---------- log
        logger.info('Remove duplicated data')
        if self.n_fittest > 0:
            logger.info(f'Survival of the fittest: top {self.n_fittest} structures survive')

    def set_tournament(self):
        '''
        setting for tournament selection
        '''
        # ---------- set self.get_parents() <-- self._tournament or self._roulette
        self.get_parents = self._tournament

    def _tournament(self, n_parent):
        '''
        tournament selection

        # ---------- args
        n_parent (int): number of parents

        # ---------- return
        parent_id (list): ID of selected parents
        '''
        parent_id = []
        while len(parent_id) < n_parent:
            t_indx = np.random.choice(len(self.ranking_dedupe), rin.t_size,
                                      replace=False)
            if parent_id:    # not allow the same parent in crossover
                if parent_id[0] == self.ranking_dedupe[min(t_indx)]:
                    continue
            parent_id.append(self.ranking_dedupe[min(t_indx)])
        return parent_id

    def set_roulette(self):
        '''
        setting for roulette selection
        '''
        # ---------- calculate cumulative fitness
        fitness_dedupe = np.array(
            [self.fitness[i] for i in self.ranking_dedupe])
        fitness_dedupe = self._linear_scaling(fitness_dedupe)
        self.cum_fit = np.cumsum(fitness_dedupe/fitness_dedupe.sum())
        # ---------- set self.get_parents <-- self._tournament or self._roulette
        self.get_parents = self._roulette

    def _roulette(self, n_parent):
        '''
        roulette selection

        # ---------- args
        n_parent (int): number of parents

        # ---------- return
        parent_id (list): ID of selected parents
        '''
        # ---------- select parents
        parent_id = []
        while len(parent_id) < n_parent:
            # indx_array: if all false, array is vacant
            indx_array = np.where(self.cum_fit < np.random.rand())[0]
            # select_indx: consider vacant or not
            select_indx = indx_array[-1] + 1 if indx_array.size != 0 else 0
            if parent_id:    # not allow same parent for crossover
                if parent_id[0] == self.ranking_dedupe[select_indx]:
                    continue
            parent_id.append(self.ranking_dedupe[select_indx])
        return parent_id

    def _linear_scaling(self, fitness):
        # ---------- check args
        # ------ fitness
        if isinstance(fitness, np.ndarray):
            pass
        elif isinstance(fitness, list):
            fitness = np.array(fitness)
        else:
            logger.error('fitness must be list or np.array')
        # ------ in case of int
        a = float(rin.a_rlt)
        b = float(rin.b_rlt)
        # ---------- for fit_reverse
        if not rin.fit_reverse:
            fitness = -fitness
        # ---------- scaling
        fmax = float(fitness.max())
        fmin = float(fitness.min())
        # ------ in case the same values
        if fmax == fmin:
            return fitness
        fitness = (a - b)/(fmax - fmin)*fitness + (
            (b*fmax - a*fmin)/(fmax - fmin))
        # ---------- return
        return fitness



'''
Select an optimizer
    - VASP
    - QE
    - OMX
    - soiap
    - LAMMPS
    - ASE
'''

from logging import getLogger

from ..IO import read_input as rin

if rin.calc_code == 'VASP':
    from .VASP import calc_files_vasp, ctrl_job_vasp, collect_vasp
elif rin.calc_code == 'QE':
    from .QE import calc_files_qe, ctrl_job_qe, collect_qe
elif rin.calc_code == 'OMX':
    from .OMX import calc_files_OMX, ctrl_job_OMX, collect_OMX
elif rin.calc_code == 'soiap':
    from .soiap import calc_files_soiap, ctrl_job_soiap, collect_soiap
elif rin.calc_code == 'LAMMPS':
    from .LAMMPS import calc_files_lammps, ctrl_job_lammps, collect_lammps
elif rin.calc_code == 'ASE':
    from .ASE import calc_files_ase, ctrl_job_ase, collect_ase
elif rin.calc_code == 'ext':
    from .ext import collect_ext


logger = getLogger('cryspy')

def check_calc_files():
    if rin.calc_code == 'VASP':
        calc_files_vasp.check_input_vasp()
    elif rin.calc_code == 'QE':
        calc_files_qe.check_input_qe()
    elif rin.calc_code == 'OMX':
        calc_files_OMX.check_input_OMX()
    elif rin.calc_code == 'soiap':
        calc_files_soiap.check_input_soiap()
    elif rin.calc_code == 'LAMMPS':
        calc_files_lammps.check_input_lammps()
    elif rin.calc_code == 'ASE':
        calc_files_ase.check_input_ase()
    elif rin.calc_code == 'ext':
        pass
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def next_stage(stage, work_path, *args):
    # args[0] <-- kpt_data
    # args[1] <-- current_id for kpt_data
    if rin.calc_code == 'VASP':
        skip_flag, kpt_data = ctrl_job_vasp.next_stage_vasp(stage, work_path,
                                                            args[0], args[1])
        return skip_flag, kpt_data
    elif rin.calc_code == 'QE':
        skip_flag, kpt_data = ctrl_job_qe.next_stage_qe(stage, work_path,
                                                        args[0], args[1])
        return skip_flag, kpt_data
    elif rin.calc_code == 'OMX':
        skip_flag, kpt_data = ctrl_job_OMX.next_stage_OMX(stage, work_path,
                                                          args[0], args[1])
        return skip_flag, kpt_data
    elif rin.calc_code == 'soiap':
        skip_flag = ctrl_job_soiap.next_stage_soiap(stage, work_path)
        return skip_flag
    elif rin.calc_code == 'LAMMPS':
        skip_flag = ctrl_job_lammps.next_stage_lammps(stage, work_path)
        return skip_flag
    elif rin.calc_code == 'ASE':
        skip_flag = ctrl_job_ase.next_stage_ase(stage, work_path)
        return skip_flag
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def collect(current_id, work_path):
    if rin.calc_code == 'VASP':
        opt_struc, energy, magmom, check_opt = \
            collect_vasp.collect_vasp(current_id, work_path)
    elif rin.calc_code == 'QE':
        opt_struc, energy, magmom, check_opt = \
            collect_qe.collect_qe(current_id, work_path)
    elif rin.calc_code == 'OMX':
        opt_struc, energy, magmom, check_opt = \
            collect_OMX.collect_OMX(current_id, work_path)
    elif rin.calc_code == 'soiap':
        opt_struc, energy, magmom, check_opt = \
            collect_soiap.collect_soiap(current_id, work_path)
    elif rin.calc_code == 'LAMMPS':
        opt_struc, energy, magmom, check_opt = \
            collect_lammps.collect_lammps(current_id, work_path)
    elif rin.calc_code == 'ASE':
        opt_struc, energy, magmom, check_opt = \
            collect_ase.collect_ase(current_id, work_path)
    elif rin.calc_code == 'ext':
        ext_opt_struc_data, ext_energy_data = collect_ext.collect_ext()
        return ext_opt_struc_data, ext_energy_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)

    # ---------- return
    return opt_struc, energy, magmom, check_opt


def next_struc(structure, current_id, work_path, *args):
    # args[0] <-- kpt_data
    if rin.calc_code == 'VASP':
        kpt_data = ctrl_job_vasp.next_struc_vasp(structure, current_id,
                                                 work_path, args[0])
        return kpt_data
    elif rin.calc_code == 'QE':
        kpt_data = ctrl_job_qe.next_struc_qe(structure, current_id,
                                             work_path, args[0])
        return kpt_data
    elif rin.calc_code == 'OMX':
        kpt_data = ctrl_job_OMX.next_struc_OMX(structure, current_id,
                                               work_path, args[0])
        return kpt_data
    elif rin.calc_code == 'soiap':
        ctrl_job_soiap.next_struc_soiap(structure, current_id, work_path)
    elif rin.calc_code == 'LAMMPS':
        ctrl_job_lammps.next_struc_lammps(structure, current_id, work_path)
    elif rin.calc_code == 'ASE':
        ctrl_job_ase.next_struc_ase(structure, current_id, work_path)
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def get_energy_step(energy_step_data, current_id, work_path):
    if rin.calc_code == 'VASP':
        energy_step_data = collect_vasp.get_energy_step_vasp(
            energy_step_data, current_id, work_path)
        return energy_step_data
    elif rin.calc_code == 'QE':
        energy_step_data = collect_qe.get_energy_step_qe(energy_step_data,
                                                         current_id, work_path)
        return energy_step_data
    elif rin.calc_code == 'soiap':
        energy_step_data = collect_soiap.get_energy_step_soiap(
            energy_step_data, current_id, work_path)
        return energy_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def get_struc_step(struc_step_data, current_id, work_path):
    if rin.calc_code == 'VASP':
        struc_step_data = collect_vasp.get_struc_step_vasp(
            struc_step_data, current_id, work_path)
        return struc_step_data
    elif rin.calc_code == 'QE':
        struc_step_data = collect_qe.get_struc_step_qe(
            struc_step_data, current_id, work_path)
        return struc_step_data
    elif rin.calc_code == 'soiap':
        struc_step_data = collect_soiap.get_struc_step_soiap(
            struc_step_data, current_id, work_path)
        return struc_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def get_force_step(force_step_data, current_id, work_path):
    if rin.calc_code == 'VASP':
        force_step_data = collect_vasp.get_force_step_vasp(
            force_step_data, current_id, work_path)
        return force_step_data
    elif rin.calc_code == 'QE':
        force_step_data = collect_qe.get_force_step_qe(
            force_step_data, current_id, work_path)
        return force_step_data
    elif rin.calc_code == 'soiap':
        force_step_data = collect_soiap.get_force_step_soiap(
            force_step_data, current_id, work_path)
        return force_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)


def get_stress_step(stress_step_data, current_id, work_path):
    if rin.calc_code == 'VASP':
        stress_step_data = collect_vasp.get_stress_step_vasp(
            stress_step_data, current_id, work_path)
        return stress_step_data
    elif rin.calc_code == 'QE':
        stress_step_data = collect_qe.get_stress_step_qe(
            stress_step_data, current_id, work_path)
        return stress_step_data
    elif rin.calc_code == 'soiap':
        stress_step_data = collect_soiap.get_stress_step_soiap(
            stress_step_data, current_id, work_path)
        return stress_step_data
    else:
        logger.error(f'{rin.calc_code}: not implemented yet')
        raise SystemExit(1)



'''
Control jobs
'''

import itertools
from logging import getLogger
import os
import shutil
import subprocess

import numpy as np

from ..interface import select_code
from ..IO import read_input as rin
from ..IO import change_input, io_stat, pkl_data
from ..IO.out_results import out_rslt
from ..util.utility import backup_cryspy
from ..util.struc_util import out_poscar, out_cif

if rin.algo == 'BO':
    from ..BO.select_descriptor import select_descriptor
    from ..BO import bo_next_select
if rin.algo == 'EA':
    from ..EA import ea_next_gen
if rin.algo == 'LAQA':
    from ..LAQA.calc_score import calc_laqa_bias
    from ..LAQA import laqa_next_selection
    from ..IO.out_results import out_laqa_status, out_laqa_step, out_laqa_score
    from ..IO.out_results import out_laqa_energy, out_laqa_bias


logger = getLogger('cryspy')

class Ctrl_job:

    def __init__(self, stat, init_struc_data):
        self.stat = stat
        self.init_struc_data = init_struc_data
        self.opt_struc_data = pkl_data.load_opt_struc()
        self.rslt_data = pkl_data.load_rslt()
        self.recheck = False
        self.logic_next = False
        # ---------- for each algorithm
        if rin.algo == 'RS':
            self.id_queueing, self.id_running = pkl_data.load_rs_id()
        elif rin.algo == 'BO':
            (self.n_selection, self.id_queueing,
             self.id_running, self.id_select_hist) = pkl_data.load_bo_id()
            (self.init_dscrpt_data, self.opt_dscrpt_data,
             self.bo_mean, self.bo_var,
             self.bo_score) = pkl_data.load_bo_data()
        elif rin.algo == 'LAQA':
            (self.id_queueing, self.id_running,
             self.id_select_hist) = pkl_data.load_laqa_id()
            (self.tot_step_select, self.laqa_step,
             self.laqa_struc, self.laqa_energy,
             self.laqa_bias, self.laqa_score) = pkl_data.load_laqa_data()
        elif rin.algo == 'EA':
            (self.gen, self.id_queueing,
             self.id_running) = pkl_data.load_ea_id()
            if rin.struc_mode in ['mol', 'mol_bs']:
                self.struc_mol_id = pkl_data.load_struc_mol_id()
            # do not have to load ea_data here.
            # ea_data is used only in ea_next_gen.py
        # ---------- for option
        if rin.kpt_flag:
            self.kpt_data = pkl_data.load_kpt()
        if rin.energy_step_flag:
            self.energy_step_data = pkl_data.load_energy_step()
        if rin.struc_step_flag:
            self.struc_step_data = pkl_data.load_struc_step()
        if rin.force_step_flag:
            self.force_step_data = pkl_data.load_force_step()
        if rin.stress_step_flag:
            self.stress_step_data = pkl_data.load_stress_step()
        # ---------- flag for next selection or generation
        if rin.algo in ['BO', 'LAQA', 'EA']:
            if (self.id_queueing or self.id_running):
                self.go_next_sg = False
            else:
                self.go_next_sg = True

    def check_job(self):
        # ---------- option: recalc
        if rin.recalc:
            self.set_recalc()
        # ---------- temporarily append
        self.tmp_running = self.id_running[:]    # shallow copy
        self.tmp_queueing = self.id_queueing[:]
        if not rin.stop_next_struc:    # if true --> option: stop_next_struc
            while len(self.tmp_running) < rin.njob and self.tmp_queueing:
                self.tmp_running.append(self.tmp_queueing.pop(0))
        # ---------- initialize
        self.stage_stat = {}    # key: Structure ID
        self.job_stat = {}
        # ---------- check job status
        for cid in self.tmp_running:
            # ------ mkdir
            if not os.path.isdir(f'work/{cid:06}'):
                os.mkdir(f'work/{cid:06}')
            # ------ check stat_job file
            stat_path = f'work/{cid:06}' + '/stat_job'
            try:
                with open(stat_path, 'r') as fstat:
                    istat = fstat.readline()    # id
                    sstat = fstat.readline()    # stage
                    jstat = fstat.readline()    # submitted or done or ...
                self.stage_stat[cid] = int(sstat.split()[0])
                if not cid == int(istat.split()[0]):
                    logger.error(f'ID is wrong in work/{cid:06}')
                    raise SystemExit(1)
                self.stage_stat[cid] = int(sstat.split()[0])
                if jstat[0:3] == 'sub':
                    self.job_stat[cid] = 'submitted'
                elif jstat[0:4] == 'done':
                    self.job_stat[cid] = 'done'
                elif jstat[0:4] == 'skip':
                    self.job_stat[cid] = 'skip'
                else:
                    self.job_stat[cid] = 'else'
            except IOError:
                self.stage_stat[cid] = 'no_file'
                self.job_stat[cid] = 'no_file'

    def set_recalc(self):
        # ---------- check id
        for tid in rin.recalc:
            if tid not in self.opt_struc_data:
                logger.error(f'ID {tid} has not yet been calculated')
                raise SystemExit(1)
        # ---------- append IDs to the head of id_queueing
        self.id_queueing = rin.recalc + self.id_queueing
        io_stat.set_id(self.stat, 'id_queueing', self.id_queueing)
        self.save_id_data()
        # ---------- log and out
        logger.info('# -- Recalc')
        logger.info(f'Append {rin.recalc} to the head of id_queueing')
        # ---------- clear recalc
        rin.recalc = []
        config = change_input.config_read()
        change_input.change_option(config, 'recalc', '')    # clear
        change_input.write_config(config)
        logger.info('Clear recalc in cryspy.in')
        io_stat.set_input_common(self.stat, 'option', 'recalc', '')
        io_stat.write_stat(self.stat)

    def handle_job(self):
        logger.info('# ---------- job status')
        for cid in self.tmp_running:
            # ---------- set work_path and current_id
            self.work_path = f'./work/{cid:06}/'
            self.current_id = cid
            # ---------- handle job
            if self.job_stat[cid] == 'submitted':
                logger.info(f'ID {cid:>6}: still queueing or running')
            elif self.job_stat[cid] == 'done':
                self.ctrl_done()
            elif self.job_stat[cid] == 'skip':
                self.ctrl_skip()
            elif self.job_stat[cid] == 'else':
                logger.error(f'Wrong job_stat in {self.work_path}. ')
            elif self.job_stat[cid] == 'no_file':
                self.ctrl_next_struc()
            else:
                logger.error(f'Unexpected error in {self.work_path}stat_job')

    def ctrl_done(self):
        self.current_stage = self.stage_stat[self.current_id]
        # ---------- log
        logger.info(f'ID {self.current_id:>6}: Stage {self.current_stage} Done!')
        # ---------- next stage
        if self.current_stage < rin.nstage:
            self.ctrl_next_stage()
        # ---------- collect result
        elif self.current_stage == rin.nstage:
            self.ctrl_collect()
        # ---------- error
        else:
            logger.error('Wrong stage in '+self.work_path+'stat_job')
            raise SystemExit(1)

    def ctrl_next_stage(self):
        # ---------- energy step
        if rin.energy_step_flag:
            self.energy_step_data = select_code.get_energy_step(
                self.energy_step_data, self.current_id, self.work_path)
        # ---------- struc step
        if rin.struc_step_flag:
            self.struc_step_data = select_code.get_struc_step(
                self.struc_step_data, self.current_id, self.work_path)
        # ---------- force step
        if rin.force_step_flag:
            self.force_step_data = select_code.get_force_step(
                self.force_step_data, self.current_id, self.work_path)
        # ---------- stress step
        if rin.stress_step_flag:
            self.stress_step_data = select_code.get_stress_step(
                self.stress_step_data, self.current_id, self.work_path)
        # ---------- next stage
        if rin.kpt_flag:
            skip_flag, self.kpt_data = select_code.next_stage(
                self.current_stage, self.work_path,
                self.kpt_data, self.current_id)
        else:
            skip_flag = select_code.next_stage(self.current_stage,
                                               self.work_path)
        # ---------- skip
        if skip_flag:
            self.ctrl_skip()
            return
        # ---------- prepare jobfile
        self.prepare_jobfile()
        # ---------- submit
        self.submit_next_stage()

    def submit_next_stage(self):
        # ---------- submit job
        os.chdir(self.work_path)    # cd work_path
        with open('stat_job', 'w') as fwstat:
            fwstat.write(f'{self.current_id:<6}    # Structure ID\n')
            fwstat.write(f'{self.current_stage + 1:<6}    # Stage\n')
            fwstat.write('submitted\n')
        with open('sublog', 'w') as logf:
            subprocess.Popen([rin.jobcmd, rin.jobfile],
                             stdout=logf, stderr=logf)
        os.chdir('../../')    # go back to ..
        # ---------- save status
        io_stat.set_stage(self.stat, self.current_id, self.current_stage + 1)
        io_stat.write_stat(self.stat)
        # ---------- log
        logger.info(f'    submitted job, ID {self.current_id:>6} Stage {self.current_stage + 1}')

    def ctrl_collect(self):
        # ---------- energy step
        if rin.energy_step_flag:
            self.energy_step_data = select_code.get_energy_step(
                self.energy_step_data, self.current_id, self.work_path)
        # ---------- struc step
        if rin.struc_step_flag:
            self.struc_step_data = select_code.get_struc_step(
                self.struc_step_data, self.current_id, self.work_path)
        # ---------- force step
        if rin.force_step_flag:
            self.force_step_data = select_code.get_force_step(
                self.force_step_data, self.current_id, self.work_path)
        # ---------- stress step
        if rin.stress_step_flag:
            self.stress_step_data = select_code.get_stress_step(
                self.stress_step_data, self.current_id, self.work_path)
        # ---------- each algo
        if rin.algo == 'RS':
            self.ctrl_collect_rs()
        elif rin.algo == 'BO':
            self.ctrl_collect_bo()
        elif rin.algo == 'LAQA':
            self.ctrl_collect_laqa()
        elif rin.algo == 'EA':
            self.ctrl_collect_ea()
        else:
            logger.error('Error, algo')
            raise SystemExit(1)
        # ---------- move to fin
        if rin.algo == 'LAQA':
            if self.fin_laqa:
                self.mv_fin()
            else:
                os.rename(self.work_path+'stat_job',
                          self.work_path+'prev_stat_job')
        else:
            self.mv_fin()
        # ---------- update status
        self.update_status(operation='fin')
        # ---------- recheck
        self.recheck = True

    def ctrl_collect_rs(self):
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.current_id, self.work_path)
        logger.info(f'    collect results: E = {energy} eV/atom')
        # ---------- register opt_struc
        spg_sym, spg_num, spg_sym_opt, spg_num_opt = self.regist_opt(opt_struc)
        # ---------- save rslt
        self.rslt_data.loc[self.current_id] = [spg_num, spg_sym,
                                               spg_num_opt, spg_sym_opt,
                                               energy, magmom, check_opt]
        pkl_data.save_rslt(self.rslt_data)
        out_rslt(self.rslt_data)

    def ctrl_collect_bo(self):
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.current_id, self.work_path)
        logger.info(f'    collect results: E = {energy} eV/atom')
        # ---------- register opt_struc
        spg_sym, spg_num, spg_sym_opt, spg_num_opt = self.regist_opt(opt_struc)
        # ---------- save rslt
        self.rslt_data.loc[self.current_id] = [self.n_selection,
                                               spg_num, spg_sym,
                                               spg_num_opt, spg_sym_opt,
                                               energy, magmom, check_opt]
        pkl_data.save_rslt(self.rslt_data)
        out_rslt(self.rslt_data)
        # ---------- success
        if opt_struc is not None:
            # ------ calc descriptor for opt sturcture
            tmp_dscrpt = select_descriptor({self.current_id: opt_struc})
            # ------ update descriptors
            self.opt_dscrpt_data.update(tmp_dscrpt)
        # ---------- error
        else:
            # ------ update descriptors and non_error_id
            self.opt_dscrpt_data[self.current_id] = None
        # ---------- save bo_data
        self.save_data()

    def ctrl_collect_laqa(self):
        # ---------- flag for finish
        self.fin_laqa = False
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.current_id, self.work_path)
        # ---------- total step and laqa_step
        #     force_step_data[ID][stage][step][atom]
        if self.force_step_data[self.current_id][-1] is None:
            self.laqa_step[self.current_id].append(0)
        else:
            self.tot_step_select[-1] += len(
                self.force_step_data[self.current_id][-1])
            self.laqa_step[self.current_id].append(
                len(self.force_step_data[self.current_id][-1]))
        # ------ save status
        io_stat.set_common(self.stat, 'total_step', sum(self.tot_step_select))
        io_stat.write_stat(self.stat)
        # ---------- append laqa struc
        self.laqa_struc[self.current_id].append(opt_struc)
        # ---------- append laqa energy
        self.laqa_energy[self.current_id].append(energy)
        # ---------- append laqa bias
        #     force_step_data[ID][stage][step][atom]
        #     stress_step_data[ID][stage][step][atom]
        tmp_laqa_bias = calc_laqa_bias(
            self.force_step_data[self.current_id][-1],
            self.stress_step_data[self.current_id][-1],
            wf=rin.wf, ws=rin.ws)
        self.laqa_bias[self.current_id].append(tmp_laqa_bias)
        # ---------- append laqa score
        if check_opt == 'done' or np.isnan(energy) or np.isnan(tmp_laqa_bias):
            self.laqa_score[self.current_id].append(-float('inf'))
        else:
            self.laqa_score[self.current_id].append(-energy + tmp_laqa_bias)
        # ---------- save and out laqa data
        self.save_data()
        out_laqa_status(self.laqa_step, self.laqa_score,
                        self.laqa_energy, self.laqa_bias)
        out_laqa_step(self.laqa_step)
        out_laqa_score(self.laqa_score)
        out_laqa_energy(self.laqa_energy)
        out_laqa_bias(self.laqa_bias)
        # ---------- case of 'done' or error
        if check_opt == 'done' or np.isnan(energy) or np.isnan(tmp_laqa_bias):
            self.fin_laqa = True
            logger.info(f'    collect results: E = {energy} eV/atom')
            # ------ register opt_struc
            (spg_sym, spg_num,
             spg_sym_opt, spg_num_opt) = self.regist_opt(opt_struc)
            # ------ save rslt
            self.rslt_data.loc[self.current_id] = [spg_num, spg_sym,
                                                   spg_num_opt, spg_sym_opt,
                                                   energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)

    def ctrl_collect_ea(self):
        # ---------- get opt data
        opt_struc, energy, magmom, check_opt = \
            select_code.collect(self.current_id, self.work_path)
        logger.info(f'    collect results: E = {energy} eV/atom')
        # ---------- register opt_struc
        spg_sym, spg_num, spg_sym_opt, spg_num_opt = self.regist_opt(opt_struc)
        # ---------- save rslt
        self.rslt_data.loc[self.current_id] = [self.gen,
                                               spg_num, spg_sym,
                                               spg_num_opt, spg_sym_opt,
                                               energy, magmom, check_opt]
        pkl_data.save_rslt(self.rslt_data)
        out_rslt(self.rslt_data)
        # ------ success
        # if opt_struc is not None:
        #    self.save_id_data()

    def regist_opt(self, opt_struc):
        '''
        Common part in ctrl_collect_*
        '''
        # ---------- get initial spg info
        try:
            spg_sym, spg_num = self.init_struc_data[
                self.current_id].get_space_group_info(symprec=rin.symprec)
        except TypeError:
            spg_num = 0
            spg_sym = None
        # ---------- success
        if opt_struc is not None:
            # ------ get opt spg info
            try:
                spg_sym_opt, spg_num_opt = opt_struc.get_space_group_info(
                    symprec=rin.symprec)
            except TypeError:
                spg_num_opt = 0
                spg_sym_opt = None
            # ------ out opt_struc
            out_poscar(opt_struc, self.current_id, './data/opt_POSCARS')
            try:
                out_cif(opt_struc, self.current_id, self.work_path,
                        './data/opt_CIFS.cif', rin.symprec)
            except TypeError:
                logger.warning('failed to write opt_CIF')
        # ---------- error
        else:
            spg_num_opt = 0
            spg_sym_opt = None
        # ---------- register opt_struc
        self.opt_struc_data[self.current_id] = opt_struc
        pkl_data.save_opt_struc(self.opt_struc_data)
        # ---------- return
        return spg_sym, spg_num, spg_sym_opt, spg_num_opt

    def ctrl_next_struc(self):
        # ---------- RS
        if rin.algo == 'RS':
            next_struc_data = self.init_struc_data[self.current_id]
        # ---------- BO
        elif rin.algo == 'BO':
            next_struc_data = self.init_struc_data[self.current_id]
        # ---------- LAQA
        elif rin.algo == 'LAQA':
            if self.laqa_struc[self.current_id]:    # vacant list?
                next_struc_data = self.laqa_struc[self.current_id][-1]
            else:
                next_struc_data = self.init_struc_data[self.current_id]
        # ---------- EA
        elif rin.algo == 'EA':
            next_struc_data = self.init_struc_data[self.current_id]
        # ---------- algo is wrong
        else:
            logger.error('Error, algo in ctrl_next_struc')
            raise SystemExit(1)
        # ---------- common part
        # ------ in case there is no initial strucure data
        if next_struc_data is None:
            logger.info(f'ID {self.current_id:>6}: initial structure is None')
            self.ctrl_skip()
        # ------ normal initial structure data
        else:
            # -- prepare input files for structure optimization
            if rin.kpt_flag:
                self.kpt_data = select_code.next_struc(next_struc_data,
                                                       self.current_id,
                                                       self.work_path,
                                                       self.kpt_data)
            else:
                select_code.next_struc(next_struc_data, self.current_id,
                                       self.work_path)
            # -- prepare jobfile
            self.prepare_jobfile()
            # -- submit
            self.submit_next_struc()
            logger.info(f'ID {self.current_id:>6}: submit job, Stage 1')
            # -- update status
            self.update_status(operation='submit')

    def submit_next_struc(self):
        # ---------- submit job
        os.chdir(self.work_path)    # cd work_path
        with open('stat_job', 'w') as fwstat:
            fwstat.write(f'{self.current_id:<6}    # Structure ID\n')
            fwstat.write(f'{1:<6}    # Stage\n')
            fwstat.write('submitted\n')
        with open('sublog', 'w') as logf:
            subprocess.Popen([rin.jobcmd, rin.jobfile],
                             stdout=logf, stderr=logf)
        os.chdir('../../')    # go back to csp root dir

    def ctrl_skip(self):
        # ---------- log
        logger.info(f'ID {self.current_id:>6}: Skip')
        # ---------- get initial spg info
        if self.init_struc_data[self.current_id] is None:
            spg_sym = None
            spg_num = 0
        else:
            try:
                spg_sym, spg_num = self.init_struc_data[
                    self.current_id].get_space_group_info(symprec=rin.symprec)
            except TypeError:
                spg_num = 0
                spg_sym = None
        # ---------- 'skip' for rslt
        spg_num_opt = 0
        spg_sym_opt = None
        energy = np.nan
        magmom = np.nan
        check_opt = 'skip'
        # ---------- register opt_struc
        self.opt_struc_data[self.current_id] = None
        pkl_data.save_opt_struc(self.opt_struc_data)
        # ---------- RS
        if rin.algo == 'RS':
            # ------ save rslt
            self.rslt_data.loc[self.current_id] = [spg_num, spg_sym,
                                                   spg_num_opt, spg_sym_opt,
                                                   energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
        # ---------- BO
        elif rin.algo == 'BO':
            # ------ save rslt
            self.rslt_data.loc[self.current_id] = [self.n_selection,
                                                   spg_num, spg_sym,
                                                   spg_num_opt, spg_sym_opt,
                                                   energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
            # ------ update descriptors
            self.opt_dscrpt_data[self.current_id] = None
            # ------ save
            self.save_id_data()
            self.save_data()
        # ---------- LAQA
        elif rin.algo == 'LAQA':
            # ------ save rslt
            self.rslt_data.loc[self.current_id] = [spg_num, spg_sym,
                                                   spg_num_opt, spg_sym_opt,
                                                   energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
            # ---------- laqa data
            self.laqa_step[self.current_id].append(0)
            self.laqa_struc[self.current_id].append(None)
            self.laqa_energy[self.current_id].append(energy)
            self.laqa_bias[self.current_id].append(np.nan)
            self.laqa_score[self.current_id].append(-float('inf'))
            # ---------- save and out laqa data
            self.save_data()
            out_laqa_status(self.laqa_step, self.laqa_score,
                            self.laqa_energy, self.laqa_bias)
            out_laqa_step(self.laqa_step)
            out_laqa_score(self.laqa_score)
            out_laqa_energy(self.laqa_energy)
            out_laqa_bias(self.laqa_bias)
        # ---------- EA
        elif rin.algo == 'EA':
            self.rslt_data.loc[self.current_id] = [self.gen,
                                                   spg_num, spg_sym,
                                                   spg_num_opt, spg_sym_opt,
                                                   energy, magmom, check_opt]
            pkl_data.save_rslt(self.rslt_data)
            out_rslt(self.rslt_data)
        # ---------- move to fin
        self.mv_fin()
        # ---------- update status
        self.update_status(operation='fin')
        # ---------- recheck
        self.recheck = True

    def update_status(self, operation):
        # ---------- update status
        if operation == 'submit':
            self.id_running.append(self.current_id)
            self.id_queueing.remove(self.current_id)
            io_stat.set_stage(self.stat, self.current_id, 1)
        elif operation == 'fin':
            if self.current_id in self.id_queueing:
                self.id_queueing.remove(self.current_id)
            if self.current_id in self.id_running:
                self.id_running.remove(self.current_id)
            io_stat.clean_id(self.stat, self.current_id)
        else:
            logger.error('operation is wrong')
            raise SystemExit(1)
        io_stat.set_id(self.stat, 'id_queueing', self.id_queueing)
        io_stat.write_stat(self.stat)
        # ---------- save id_data
        self.save_id_data()

    def prepare_jobfile(self):
        if not os.path.isfile('./calc_in/' + rin.jobfile):
            logger.error('Could not find ./calc_in' + rin.jobfile)
            raise SystemExit(1)
        with open('./calc_in/' + rin.jobfile, 'r') as f:
            lines = f.readlines()
        lines2 = []
        for line in lines:
            lines2.append(line.replace('CrySPY_ID', str(self.current_id)))
        with open(self.work_path + rin.jobfile, 'w') as f:
            f.writelines(lines2)

    def mv_fin(self):
        if not os.path.isdir(f'work/fin/{self.current_id:06}'):
            shutil.move(f'work/{self.current_id:06}', 'work/fin/')
        else:    # rename for recalc
            for i in itertools.count(1):
                if not os.path.isdir(f'work/fin/{self.current_id:06}_{i}'):
                    shutil.move(f'work/{self.current_id:06}',
                                f'work/fin/{self.current_id:06}_{i}')
                    break

    def next_sg(self, noprint=False):
        '''
        next selection or generation
        '''
        if rin.algo == 'BO':
            self.next_select_BO(noprint)
        if rin.algo == 'LAQA':
            self.next_select_LAQA()
        if rin.algo == 'EA':
            self.next_gen_EA()

    def next_select_BO(self, noprint=False):
        # ---------- log
        logger.info(f'\nDone selection {self.n_selection}')
        # ---------- done all structures
        if len(self.rslt_data) == rin.tot_struc:
            logger.info('\nDone all structures!')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- flag for next selection or generation
        if not self.go_next_sg:
            logger.info('\nBO is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- check point 3
        if rin.stop_chkpt == 3:
            logger.info('\nStop at check point 3: BO is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- max_select_bo
        if 0 < rin.max_select_bo <= self.n_selection:
            logger.info(f'\nReached max_select_bo: {rin.max_select_bo}')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- BO
        backup_cryspy()
        bo_data = (self.init_dscrpt_data, self.opt_dscrpt_data,
                   self.bo_mean, self.bo_var, self.bo_score)
        bo_id_data = (self.n_selection, self.id_queueing,
                      self.id_running, self.id_select_hist)
        bo_next_select.next_select(self.stat, self.rslt_data,
                                   bo_id_data, bo_data, noprint)

    def next_select_LAQA(self):
        # ---------- flag for next selection or generation
        if not self.go_next_sg:
            logger.info('\nLAQA is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- check point 3
        if rin.stop_chkpt == 3:
            logger.info('\nStop at check point 3: LAQA is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- selection of LAQA
        backup_cryspy()
        laqa_id_data = (self.id_queueing, self.id_running,
                        self.id_select_hist)
        laqa_data = (self.tot_step_select, self.laqa_step, self.laqa_struc,
                     self.laqa_energy, self.laqa_bias, self.laqa_score)
        laqa_next_selection.next_selection(self.stat, laqa_id_data, laqa_data)

    def next_gen_EA(self):
        # ---------- log
        logger.info(f'\nDone generation {self.gen}')
        # ---------- flag for next selection or generation
        if not self.go_next_sg:
            logger.info('\nEA is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- check point 3
        if rin.stop_chkpt == 3:
            logger.info('\nStop at check point 3: EA is ready')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- maxgen_ea
        if 0 < rin.maxgen_ea <= self.gen:
            logger.info(f'\nReached maxgen_ea: {rin.maxgen_ea}')
            os.remove('lock_cryspy')
            raise SystemExit()
        # ---------- EA
        backup_cryspy()
        ea_id_data = (self.gen, self.id_queueing, self.id_running)
        if rin.struc_mode not in ['mol', 'mol_bs']:
            ea_next_gen.next_gen(self.stat, self.init_struc_data, None,
                                 self.opt_struc_data, self.rslt_data, ea_id_data)
        else:
            ea_next_gen.next_gen(self.stat, self.init_struc_data, self.struc_mol_id,
                                 self.opt_struc_data, self.rslt_data, ea_id_data)

    def save_id_data(self):
        # ---------- save id_data
        if rin.algo == 'RS':
            rs_id_data = (self.id_queueing, self.id_running)
            pkl_data.save_rs_id(rs_id_data)
        if rin.algo == 'BO':
            bo_id_data = (self.n_selection, self.id_queueing,
                          self.id_running, self.id_select_hist)
            pkl_data.save_bo_id(bo_id_data)
        if rin.algo == 'LAQA':
            laqa_id_data = (self.id_queueing, self.id_running,
                            self.id_select_hist)
            pkl_data.save_laqa_id(laqa_id_data)
        if rin.algo == 'EA':
            ea_id_data = (self.gen, self.id_queueing, self.id_running)
            pkl_data.save_ea_id(ea_id_data)

    def save_data(self):
        # ---------- save ??_data
        if rin.algo == 'BO':
            bo_data = (self.init_dscrpt_data, self.opt_dscrpt_data,
                       self.bo_mean, self.bo_var, self.bo_score)
            pkl_data.save_bo_data(bo_data)
        if rin.algo == 'LAQA':
            laqa_data = (self.tot_step_select, self.laqa_step, self.laqa_struc,
                         self.laqa_energy, self.laqa_bias, self.laqa_score)
            pkl_data.save_laqa_data(laqa_data)
        # ea_data is used only in ea_next_gen.py



'''
Select descriptors
    - FP: FingerPrint
'''

from logging import getLogger

from ..util.utility import check_fppath
from ..calc_dscrpt.FP.calc_FP import Calc_FP
from ..IO import read_input as rin


logger = getLogger('cryspy')

def select_descriptor(struc_data):
    # ---------- fingerprint
    if rin.dscrpt == 'FP':
        logger.info('Calculate descriptors: FingerPrint')
        # ------ check cal_fingerprint executable file
        fppath = check_fppath(rin.fppath)
        # ------ calc fingerprint
        cfp = Calc_FP(struc_data, rin.fp_rmin, rin.fp_rmax,
                      rin.fp_npoints, rin.fp_sigma, fppath)
        cfp.calc()
        return cfp.descriptors
    else:
        logger.error('Now FP only')
        raise SystemExit(1)



'''
Change input variables in cryspy.in
'''

import configparser


def config_read():
    config = configparser.ConfigParser()
    config.read('cryspy.in')
    return config


def write_config(config):
    with open('cryspy.in', 'w') as f:
        config.write(f)


def change_basic(config, var_str, var):
    config.set('basic', var_str, '{}'.format(var))


def change_option(config, var_str, var):
    config.set('option', var_str, '{}'.format(var))



'''
Restart in LAQA
'''

from logging import getLogger

from ..IO import io_stat, pkl_data
from ..IO import read_input as rin


logger = getLogger('cryspy')

def restart(stat, prev_nstruc):
    # ---------- load laqa data
    id_queueing, id_running, id_select_hist = pkl_data.load_laqa_id()
    tot_step_select, laqa_step, laqa_struc, \
        laqa_energy, laqa_bias, laqa_score = pkl_data.load_laqa_data()

    # ---------- initialize for appended ID
    for i in range(prev_nstruc, rin.tot_struc):
        laqa_step[i] = []
        laqa_struc[i] = []
        laqa_energy[i] = []
        laqa_bias[i] = []
        laqa_score[i] = [float('inf')]
        id_queueing.append(i)
    logger.info('Append scores and id_queueing')

    # ---------- status
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- save for LAQA
    laqa_id_data = (id_queueing, id_running, id_select_hist)
    pkl_data.save_laqa_id(laqa_id_data)
    laqa_data = (tot_step_select, laqa_step, laqa_struc,
                 laqa_energy, laqa_bias, laqa_score)
    pkl_data.save_laqa_data(laqa_data)



'''
Restart Bayesian optimization
'''

from .select_descriptor import select_descriptor
from ..IO import pkl_data


def restart(init_struc_data, prev_nstruc):
    # ---------- load BO data
    (init_dscrpt_data, opt_dscrpt_data,
     bo_mean, bo_var, bo_score) = pkl_data.load_bo_data()

    # ---------- get additional struc data
    tmp_struc_data = {cid: init_struc_data[cid] for cid in range(
        prev_nstruc, len(init_struc_data))}

    # ---------- calc descriptor
    tmp_dscrpt = select_descriptor(tmp_struc_data)

    # ---------- update
    init_dscrpt_data.update(tmp_dscrpt)

    # ---------- save
    bo_data = (init_dscrpt_data, opt_dscrpt_data, bo_mean, bo_var, bo_score)
    pkl_data.save_bo_data(bo_data)



from logging import getLogger

from .gen_struc_EA.crossover import Crossover
from .gen_struc_EA.ea_generation import EA_generation
from .gen_struc_EA.permutation import Permutation
from .gen_struc_EA.rotation import Rotation
from .gen_struc_EA.strain import Strain
from ..IO import pkl_data
from ..IO import read_input as rin
from ..util.utility import check_fwpath
from ..util.struc_util import set_mindist

# ---------- import later
#from ..RS.gen_struc_RS.gen_pyxtal import Rnd_struc_gen_pyxtal
#from ..RS.gen_struc_RS.random_generation import Rnd_struc_gen


logger = getLogger('cryspy')

def child_gen(sp, init_struc_data, struc_mol_id=None):
    # ---------- instantiate EA_generation class
    eagen = EA_generation(sp=sp, id_start=rin.tot_struc,
                          init_pos_path='./data/init_POSCARS')
    # ---------- set mindist
    logger.info('# mindist')
    mindist = set_mindist(rin.mindist, rin.mindist_factor)
    if rin.struc_mode == 'mol_bs':
        mindist_dummy = set_mindist(rin.mindist_mol_bs, rin.mindist_mol_bs_factor)
    # ------ Crossover
    if rin.n_crsov > 0:
        if rin.struc_mode not in ['mol', 'mol_bs']:
            co = Crossover(mindist)
            eagen.gen_crossover(co, None, molecular=False)
        else:
            co = Crossover(mindist)
            eagen.gen_crossover(co, struc_mol_id, molecular=True)

    # ------ Permutation
    if rin.n_perm > 0:
        if rin.struc_mode not in ['mol', 'mol_bs']:
            pm = Permutation(mindist)
            eagen.gen_permutation(pm, None, molecular=False)
        else:
            pm = Permutation(mindist)
            eagen.gen_permutation(pm, struc_mol_id, molecular=True)

    # ------ Strain
    if rin.n_strain > 0:
        if rin.struc_mode not in ['mol', 'mol_bs']:
            st = Strain(mindist)
            eagen.gen_strain(st, None)
        else:
            st = Strain(mindist)
            eagen.gen_strain(st, struc_mol_id, protect_mol_struc=True)

    # ------ Rotation
    if rin.struc_mode in ['mol', 'mol_bs']:
        if rin.n_rotation > 0:
            rot = Rotation(mindist)
            eagen.gen_rotation(struc_mol_id, rot=rot)

    # ------ update init_struc_data
    init_struc_data.update(eagen.offspring)
    if rin.struc_mode in ['mol', 'mol_bs']:
        struc_mol_id.update(eagen.offspring_mol_id)

    # ---------- random generation
    if rin.n_rand > 0:
        # ------ pyxtal
        if not (rin.spgnum == 0 or rin.use_find_wy):
            from ..RS.gen_struc_RS.gen_pyxtal import Rnd_struc_gen_pyxtal
            rsgx = Rnd_struc_gen_pyxtal(mindist=mindist)
            # -- crystal
            if rin.struc_mode == 'crystal':
                rsgx.gen_struc(nstruc=rin.n_rand, id_offset=eagen.cid,
                               init_pos_path='./data/init_POSCARS')
            # -- molecular crystal
            elif rin.struc_mode == 'mol':
                rsgx.set_mol()
                rsgx.gen_struc_mol(nstruc=rin.n_rand, id_offset=eagen.cid,
                                   init_pos_path='./data/init_POSCARS')
            # ------ molecular crystal breaking symmetry
            elif rin.struc_mode == 'mol_bs':
                logger.info('# -- mindist_mol_bs')
                mindist_dummy = set_mindist(rin.mindist_mol_bs, rin.mindist_mol_bs_factor, dummy=True)
                rsgx.set_mol()
                rsgx.gen_struc_mol_break_sym(nstruc=rin.n_rand,
                                             mindist_dummy=mindist_dummy,
                                             id_offset=eagen.cid,
                                             init_pos_path='./data/init_POSCARS')
            # -- update
            init_struc_data.update(rsgx.init_struc_data)
            if rin.struc_mode in ['mol', 'mol_bs']:
                struc_mol_id.update(rsgx.struc_mol_id)
        # ------ Rnd_struc_gen
        else:
            from ..RS.gen_struc_RS.random_generation import Rnd_struc_gen
            rsg = Rnd_struc_gen(mindist=mindist)
            if rin.spgnum == 0:
                rsg.gen_wo_spg(nstruc=rin.n_rand, id_offset=eagen.cid,
                               init_pos_path='./data/init_POSCARS')
                init_struc_data.update(rsg.init_struc_data)
            else:
                fwpath = check_fwpath(rin.fwpath)
                rsg.gen_with_find_wy(nstruc=rin.n_rand,
                                     id_offset=eagen.cid,
                                     init_pos_path='./data/init_POSCARS',
                                     fwpath=fwpath)
                init_struc_data.update(rsg.init_struc_data)

    # ---------- save init_struc_data
    pkl_data.save_init_struc(init_struc_data)
    if rin.struc_mode in ['mol', 'mol_bs']:
        pkl_data.save_struc_mol_id(struc_mol_id)

    # ----------return
    if rin.struc_mode not in ['mol', 'mol_bs']:
        return init_struc_data, eagen
    else:
        return init_struc_data, eagen, struc_mol_id



'''
Control jobs in soiap
'''

from logging import getLogger
import os
import shutil

from . import structure as soiap_structure
from ...IO import read_input as rin


logger = getLogger('cryspy')

def next_stage_soiap(stage, work_path):
    # ---------- skip_flag
    skip_flag = False

    # ---------- rename soiap files at the current stage
    soiap_files = [rin.soiap_infile, rin.soiap_outfile, rin.soiap_cif,
                   'log.struc', 'log.tote', 'log.frc', 'log.strs']
    for f in soiap_files:
        if not os.path.isfile(work_path+f):
            logger.error('Not found '+work_path+f)
            raise SystemExit(1)
        os.rename(work_path+f, work_path+'stage{}_'.format(stage)+f)

    # ---------- copy the input file from ./calc_in for the next stage
    finfile = './calc_in/'+rin.soiap_infile+'_{}'.format(stage + 1)
    shutil.copyfile(finfile, work_path+rin.soiap_infile)

    # ---------- generate the CIF file
    try:
        with open(work_path+'stage{}_log.struc'.format(stage), 'r') as f:
            lines = f.readlines()
            lines = lines[-(rin.natot+5):]
        structure = soiap_structure.from_file(lines)
    except ValueError:
        skip_flag = True
        logger.warning('    error in soiap,  skip this structure')
        return skip_flag
    with open(work_path+'stage{}_'.format(stage)+rin.soiap_cif, 'r') as f:
        lines = f.readlines()
    title = lines[0][5:]    # string following 'data_'
    soiap_structure.write(structure,
                          work_path+rin.soiap_cif,
                          symprec=rin.symprec,
                          title=title)

    # ---------- return
    return skip_flag


def next_struc_soiap(structure, current_id, work_path):
    # ---------- copy files
    calc_inputs = [rin.soiap_infile]
    for f in calc_inputs:
        ff = f+'_1' if f == rin.soiap_infile else f
        if not os.path.isfile('./calc_in/'+ff):
            logger.error('Could not find ./calc_in/'+ff)
            raise SystemExit(1)
        shutil.copyfile('./calc_in/'+ff, work_path+f)

    # ---------- generate the CIF file
    soiap_structure.write(structure,
                          work_path+rin.soiap_cif,
                          symprec=rin.symprec,
                          title='ID_{0:d}'.format(current_id))



'''
Initialize evolutionary algorithm
'''

from logging import getLogger

import pandas as pd

from ..IO import out_results
from ..IO import io_stat, pkl_data
from ..IO import read_input as rin


logger = getLogger('cryspy')

def initialize(stat, rslt_data):
    # ---------- log
    logger.info('# ---------- Initialize evolutionary algorithm')
    logger.info('# ------ Generation 1')
    logger.info(f'{rin.tot_struc} structures by random')

    # ---------- initialize
    gen = 1
    id_queueing = [i for i in range(rin.tot_struc)]
    id_running = []
    # ------ ea_info
    ea_info = pd.DataFrame(columns=['Gen', 'Population',
                                    'Crossover', 'Permutation', 'Strain',
                                    'Random', 'Elite',
                                    'crs_lat', 'slct_func'])
    ea_info.iloc[:, 0:7] = ea_info.iloc[:, 0:7].astype(int)
    tmp_info = pd.DataFrame([[1, rin.tot_struc, 0, 0, 0, rin.tot_struc, 0,
                              rin.crs_lat, rin.slct_func]],
                            columns=ea_info.columns)
    ea_info = pd.concat([ea_info, tmp_info], axis=0, ignore_index=True)
    out_results.out_ea_info(ea_info)
    # ------ ea_origin
    ea_origin = pd.DataFrame(columns=['Gen', 'Struc_ID',
                                      'Operation', 'Parent'])
    ea_origin.iloc[:, 0:2] = ea_origin.iloc[:, 0:2].astype(int)
    for cid in range(rin.tot_struc):
        tmp_origin = pd.DataFrame([[1, cid, 'random', None]],
                                  columns=ea_origin.columns)
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ elite
    elite_struc = None
    elite_fitness = None
    # ------ rslt_data
    rslt_data['Gen'] = pd.Series(dtype=int)
    rslt_data = rslt_data[['Gen', 'Spg_num',
                           'Spg_sym', 'Spg_num_opt',
                           'Spg_sym_opt', 'E_eV_atom', 'Magmom', 'Opt']]

    # ---------- save
    ea_id_data = (gen, id_queueing, id_running)
    pkl_data.save_ea_id(ea_id_data)
    ea_data = (elite_struc, elite_fitness, ea_info, ea_origin)
    pkl_data.save_ea_data(ea_data)
    pkl_data.save_rslt(rslt_data)

    # ---------- status
    io_stat.set_common(stat, 'generation', gen)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)



'''
Calculate Fingerprint
'''

from logging import getLogger
import os
import subprocess

import numpy as np


logger = getLogger('cryspy')

class Calc_FP:
    '''
    calculate fingerprint using cal_fingerprint program

    # ---------- args
    struc_data (dict or list): structure data
        You may include None in struc_data
        if type of struc_data is list,
            struc_data is converted into dict type
            as {0: struc_data_0, 1: struc_data_1, ...}

    # ---------- instance methods
    self.calc(self)

    # ---------- comment
    descriptor data in self.descriptors
    '''

    def __init__(self, struc_data, fp_rmin=0.5, fp_rmax=5.0, fp_npoints=10,
                 fp_sigma=1.0, fppath='./cal_fingerprint'):
        # ---------- check args
        # ------ struc_data
        if type(struc_data) is dict:
            pass
        elif type(struc_data) is list:
            # -- convert to dict
            struc_data = {i: struc_data[i] for i in range(len(struc_data))}
        else:
            logger.error('Type of struc_data is wrong')
            raise SystemExit(1)
        self.struc_data = struc_data
        # ------ fp
        self.fp_rmin = fp_rmin
        self.fp_rmax = fp_rmax
        self.fp_sigma = fp_sigma
        self.fp_npoints = fp_npoints
        self.fppath = os.path.abspath(fppath)

    def calc(self):
        '''
        calculate fingerprint

        # ---------- return
        self.descriptors (dict): descriptor data
        '''
        # ---------- cd temporary directory
        if not os.path.isdir('tmp_calc_FP'):
            os.mkdir('tmp_calc_FP')
        os.chdir('tmp_calc_FP')
        # ---------- calc fingerprint
        self.descriptors = {}
        for cid, struc in self.struc_data.items():
            # ------ output POSCAR
            struc.to(fmt='poscar', filename='POSCAR')
            if not os.path.isfile('POSCAR'):
                logger.error('No POSCAR file')
                raise SystemExit(1)
            # ------ run cal_fingerprint
            with open('log_fingerprint', 'w') as logf:
                subprocess.call([self.fppath, 'POSCAR',
                                 '-rmin', '{}'.format(self.fp_rmin),
                                 '-rmax', '{}'.format(self.fp_rmax),
                                 '-npoints', '{}'.format(self.fp_npoints),
                                 '-sigma', '{}'.format(self.fp_sigma)],
                                stdout=logf, stderr=logf)
            fp = np.loadtxt('feature_ffpf.dat')
            # ------ mv xxx --> fin_xxx
            os.rename('POSCAR', 'fin_POSCAR')
            os.rename('feature_ffpf.dat', 'fin_feature_ffpf.dat')
            # ------ fp --> descriptors
            self.descriptors[cid] = fp
        # ---------- go back to ..
        os.chdir('../')



'''
Restart random search
'''

from ..IO import io_stat, pkl_data
from ..IO import read_input as rin


def restart(stat, prev_nstruc):
    # ---------- load rs data
    id_queueing, id_running = pkl_data.load_rs_id()

    # ---------- append id_queueing
    id_queueing.extend([i for i in range(prev_nstruc, rin.tot_struc)])

    # ---------- status
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- save
    rs_id_data = (id_queueing, id_running)
    pkl_data.save_rs_id(rs_id_data)

    # ---------- ext
    if rin.calc_code == 'ext':
        with open('ext/stat_job', 'w') as fstat:
            fstat.write('out\n')



'''
Random structure generation using PyXtal (https://github.com/qzhu2017/PyXtal)
'''

import collections
from contextlib import redirect_stdout, redirect_stderr
from io import StringIO
from logging import getLogger
from multiprocessing import Process, Queue
import os
import random
import sys

import numpy as np
from pymatgen.core import Structure, Molecule
from pymatgen.core.periodic_table import DummySpecie
from pyxtal import pyxtal
from pyxtal.database.collection import Collection
from pyxtal.tolerance import Tol_matrix

from ...util.struc_util import check_distance, sort_by_atype, out_poscar
from ...util.struc_util import get_atype_dummy, scale_cell_mol, rot_mat
from ...IO import read_input as rin


logger = getLogger('cryspy')

class Rnd_struc_gen_pyxtal:
    '''
    Random structure generation using pyxtal
    '''

    def __init__(self, mindist):
        self.mindist = mindist

    def set_mol(self):
        '''
        set molecule files and number of molecules
        '''
        # ----------
        mol_data = []
        pyxtal_mol_data = Collection('molecules')
        pyxtal_mol_names = list(Collection('molecules'))
        for i, mf in enumerate(rin.mol_file):
            if os.path.isfile(mf):
                mol = Molecule.from_file(mf)
            elif mf in pyxtal_mol_names:
                mol = pyxtal_mol_data[mf]
            else:
                logger.error('no molecular files')
                raise SystemExit(1)
            mol_data.append(mol)
        # ---------- self.xxx
        self.mol_data = mol_data

    def gen_struc(self, nstruc, id_offset=0, init_pos_path=None):
        '''
        Generate random structures for given space groups

        # ---------- args
        nstruc (int): number of generated structures

        id_offset (int): default: 0
                         structure ID starts from id_offset
                         e.g. nstruc = 3, id_offset = 10
                              you obtain ID 10, ID 11, ID 12

        init_pos_path (str): default: None
                             specify a path of file
                             if you write POSCAR data of init_struc_data
                             ATTENSION: data are appended to the specified file

        # ---------- comment
        generated structure data are saved in self.init_struc_data
        '''
        # ---------- initialize
        self.init_struc_data = {}
        # ---------- Tol_matrix
        tolmat = self._set_tol_mat(rin.atype, self.mindist)
        # ---------- loop for structure generattion
        while len(self.init_struc_data) < nstruc:
            # ------ spgnum --> spg
            if rin.spgnum == 'all':
                spg = random.randint(1, 230)
            else:
                spg = random.choice(rin.spgnum)
            # ------ vol_factor
            rand_vol = random.uniform(rin.vol_factor[0], rin.vol_factor[1])
            # ------ generate structure
            tmp_crystal = pyxtal()
            try:
                f = StringIO()
                with redirect_stdout(f):
                    tmp_crystal.from_random(dim=3, group=spg, species=rin.atype,
                                            numIons=rin.nat, factor=rand_vol,
                                            conventional=False, tm=tolmat)
                s = f.getvalue().rstrip()    # to delete \n
                if s:
                    logger.warning(s)
            except Exception as e:
                logger.warning(e.args[0] + f': spg = {spg} retry.')
                continue
            if tmp_crystal.valid:
                tmp_struc = tmp_crystal.to_pymatgen(resort=False)    # pymatgen Structure format
                # -- check the number of atoms
                if not self._check_nat(tmp_struc):
                    # (pyxtal 0.1.4) cryspy adopts "conventional=False",
                    #     which is better for DFT calculation
                    # pyxtal returns conventional cell, that is, too many atoms
                    tmp_struc = tmp_struc.get_primitive_structure()
                    # recheck nat
                    if not self._check_nat(tmp_struc):    # failure
                        continue
                # -- sort, just in case
                tmp_struc = sort_by_atype(tmp_struc, rin.atype)
                # -- scale volume
                if rin.vol_mu is not None:
                    vol = random.gauss(mu=rin.vol_mu, sigma=rin.vol_sigma)
                    tmp_struc.scale_lattice(volume=vol)
                # -- check minimum distance
                #    from CrySPY 0.10.4
                #    Tol_matrix is used for mindist
                #
                #if self.mindist is not None:
                #    success, mindist_ij, dist = check_distance(tmp_struc,
                #                                               self.atype,
                #                                               self.mindist)
                #    if not success:
                #        print('mindist in gen_struc: {} - {}, {}. retry.'.format(
                #            self.atype[mindist_ij[0]],
                #            self.atype[mindist_ij[1]],
                #            dist), file=sys.stderr)
                #        continue    # failure
                # -- check actual space group
                try:
                    spg_sym, spg_num = tmp_struc.get_space_group_info(
                        symprec=rin.symprec)
                except TypeError:
                    spg_num = 0
                    spg_sym = None
                # -- register the structure in pymatgen format
                cid = len(self.init_struc_data) + id_offset
                self.init_struc_data[cid] = tmp_struc
                logger.info(f'Structure ID {cid:>6} was generated.'
                      f' Space group: {spg:>3} --> {spg_num:>3} {spg_sym}')
                # -- save init_POSCARS
                if init_pos_path is not None:
                    out_poscar(tmp_struc, cid, init_pos_path)

    def gen_struc_mol(self, nstruc, id_offset=0, init_pos_path=None):
        '''
        Generate random molecular crystal structures for given space groups
        one have to run self.set_mol() in advance
        # ---------- args
        nstruc (int): number of generated structures

        id_offset (int): default: 0
                         structure ID starts from id_offset
                         e.g. nstruc = 3, id_offset = 10
                              you obtain ID 10, ID 11, ID 12

        init_pos_path (str): default: None
                             specify a path of file
                             if you write POSCAR data of init_struc_data
                             ATTENSION: data are appended to the specified file

        # ---------- comment
        generated structure data are saved in self.init_struc_data
        '''
        # ---------- initialize
        self.init_struc_data = {}
        if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
            self.struc_mol_id = {}
        # ---------- Tol_matrix
        tolmat = self._set_tol_mat(rin.atype, self.mindist)
        # ---------- loop for structure generattion
        while len(self.init_struc_data) < nstruc:
            # ------ spgnum --> spg
            if rin.spgnum == 'all':
                spg = random.randint(1, 230)
            else:
                spg = random.choice(rin.spgnum)
            rand_vol = random.uniform(rin.vol_factor[0], rin.vol_factor[1])
            # ------ generate structure
            # -- multiprocess for measures against hangup
            q = Queue()
            p = Process(target=self._mp_mc, args=(tolmat, spg, rin.nmol, rand_vol, q, rin.algo))
            p.start()
            p.join(timeout=rin.timeout_mol)
            if p.is_alive():
                p.terminate()
                p.join()
            if sys.version_info.minor >= 7:
                # Process.close() available from python 3.7
                p.close()
            if q.empty():
                logger.warning('timeout for molecular structure generation. retry.')
                continue
            else:
                # -- get struc data from _mp_mc
                if rin.algo == 'EA':
                    tmp_q = q.get()
                    tmp_struc = tmp_q[0]    # structure data
                    dums = tmp_q[1]         # dummy element
                    in_dists = tmp_q[2]     # interatomic distance in a molecule
                    mol_dists = tmp_q[3]    # distance between molecules
                else:
                    tmp_struc = q.get()
                tmp_valid = q.get()
                if tmp_struc == 'error':
                    # in case of 'error', tmp_valid <-- error message (Exception)
                    logger.warning(tmp_valid.args[0] + f': spg = {spg} retry.')
                    continue
            if tmp_valid:
                # -- scale volume
                if rin.vol_mu is not None:
                    vol = random.gauss(mu=rin.vol_mu, sigma=rin.vol_sigma)
                    vol = vol * tmp_struc.num_sites / rin.natot    # for conv. cell
                    tmp_struc = scale_cell_mol(tmp_struc, self.mol_data, vol)
                    if not tmp_struc:    # case: scale_cell_mol returns False
                        logger.warning('failed scale cell. retry.')
                        continue
                # -- check nat
                if not self._check_nat(tmp_struc):
                    # cryspy adopts conventional=True
                    # pyxtal returns conventional cell,
                    # too many atoms if centering
                    tmp_struc = tmp_struc.get_primitive_structure()
                    # recheck nat
                    if not self._check_nat(tmp_struc):    # failure
                        logger.warning('different num. of atoms. retry.')
                        continue
                # -- grouping atoms for molecule using interatomic distance
                if rin.algo == 'EA':
                    loopcnt = 0
                    while loopcnt < 10:
                        loopcnt += 1
                        mol_group = []
                        group_id = 0
                        # check atom distance from dummy atom
                        for i, ts in enumerate(tmp_struc.species):
                            for dn, dummyatom in enumerate(dums):
                                if ts == dummyatom:
                                    mol_indx = int(ts.symbol[1:2])
                                    for j, tc_fc in enumerate(tmp_struc.frac_coords):
                                        if i == j:
                                            continue
                                        dist = tmp_struc.get_distance(i, j)
                                        for dist_indx, true_dist in enumerate(in_dists[dn]):
                                            if round(true_dist, 2) == round(dist, 2):
                                                mol_group.append([tc_fc, group_id, mol_indx, dist_indx])
                                                break
                                    group_id += 1
                        success_cnt = 0
                        # check number of atom per group
                        # if failed, shuffle struc data, and continue(max 10times)
                        mol_cnt = collections.Counter([x[2] for x in mol_group])
                        for mc_key, mc_value in mol_cnt.items():
                            if mc_value % len(self.mol_data[mc_key]) == 0:
                                success_cnt += 1
                        if success_cnt == len(self.mol_data):
                            break
                        random.shuffle(tmp_struc)
                    if loopcnt == 10:
                        continue
                    tmp_struc.remove_species(dums)
                # -- sort, necessary in molecular crystal
                tmp_struc = sort_by_atype(tmp_struc, rin.atype)
                # -- record group_id, mol_id, and dist_index
                if rin.algo == 'EA':
                    tmp_id = []
                    tmp_mol_indx = []
                    tmp_dist_indx = []
                    for atom_fc in tmp_struc.frac_coords:
                        for gatom in mol_group:
                            if np.array_equal(atom_fc, gatom[0]):
                                tmp_id.append(gatom[1])
                                tmp_mol_indx.append(gatom[2])
                                tmp_dist_indx.append(gatom[3])
                # -- check minimum distance
                #    from CrySPY 0.10.4
                #    Tol_matrix is used for mindist
                #
                #if self.mindist is not None:
                #    success, mindist_ij, dist = check_distance(tmp_struc,
                #                                               self.atype,
                #                                               self.mindist)
                #    if not success:
                #        print('mindist in gen_struc_mol: {} - {}, {}. retry.'.format(
                #            self.atype[mindist_ij[0]],
                #            self.atype[mindist_ij[1]],
                #            dist), file=sys.stderr)
                #        continue    # failure
                # -- check actual space group
                try:
                    spg_sym, spg_num = tmp_struc.get_space_group_info(
                        symprec=rin.symprec)
                except TypeError:
                    spg_num = 0
                    spg_sym = None
                # -- register the structure in pymatgen format
                cid = len(self.init_struc_data) + id_offset
                self.init_struc_data[cid] = tmp_struc
                if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
                    self.struc_mol_id.update({cid: [tmp_mol_indx, tmp_id, mol_dists]})
                logger.info(f'Structure ID {cid:>6} was generated.'
                      f' Space group: {spg:>3} --> {spg_num:>3} {spg_sym}')
                # -- save init_POSCARS
                if init_pos_path is not None:
                    out_poscar(tmp_struc, cid, init_pos_path)

    def gen_struc_mol_break_sym(self, nstruc, mindist_dummy,
                                id_offset=0, init_pos_path=None):
        '''
        Generate random molecular crystal structures
        one have to run self.set_mol() in advance
        molecules are put a Wyckoff position without consideration of symmetry

        # ---------- args
        nstruc (int): number of generated structures

        mindist_dummy: mindist for dummy atoms

        id_offset (int): default: 0
                         structure ID starts from id_offset
                         e.g. nstruc = 3, id_offset = 10
                              you obtain ID 10, ID 11, ID 12

        init_pos_path (str): default: None
                             specify a path of file
                             if you write POSCAR data of init_struc_data
                             ATTENSION: data are appended to the specified file

        # ---------- comment
        generated structure data are saved in self.init_struc_data
        '''
        # ---------- initialize
        self.init_struc_data = {}
        if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
            self.struc_mol_id = {}
        # ------ dummy atom type
        atype_dummy = get_atype_dummy()
        # ---------- Tol_matrix for dummy atoms
        tolmat = self._set_tol_mat(atype_dummy, mindist_dummy)
        # ---------- loop for structure generattion
        while len(self.init_struc_data) < nstruc:
            # ------ spgnum --> spg
            if rin.spgnum == 'all':
                spg = random.randint(1, 230)
            else:
                spg = random.choice(rin.spgnum)
            # ------ vol_factor
            rand_vol = random.uniform(rin.vol_factor[0], rin.vol_factor[1])
            # ------ generate structure
            tmp_crystal = pyxtal()
            try:
                f = StringIO()
                with redirect_stdout(f):
                    tmp_crystal.from_random(dim=3, group=spg, species=atype_dummy,
                                            numIons=rin.nmol, factor=rand_vol,
                                            conventional=False, tm=tolmat)
                s = f.getvalue().rstrip()    # to delete \n
                if s:
                    logger.warning(s)
            except Exception as e:
                logger.warning(e.args[0] + f': spg = {spg} retry.')
                continue
            if tmp_crystal.valid:
                # -- each wyckoff position --> dummy atom
                dums = []        # dummy atoms
                dum_pos = []     # internal position of dummy
                dum_type = {}    # type of dummy
                                 #  e.g. {DummySpecie X00+: 'Rn',
                                 #        DummySpecie X10+: 'Rn',
                                 #        DummySpecie X20+: 'Xe',
                                 #        DummySpecie X30+: 'Xe'}
                for i, site in enumerate(tmp_crystal.atom_sites):
                    dum = DummySpecie("X{}".format(i))
                    dums.append(dum)
                    dum_pos.append(site.position)
                    dum_type[dum] = site.specie
                # -- tmp_struc with dummy atoms
                lattice = tmp_crystal.lattice.get_matrix()
                tmp_struc = Structure.from_spacegroup(spg, lattice, dums, dum_pos)
                tmp_struc = tmp_struc.get_primitive_structure()
                # -- scale volume
                if rin.vol_mu is not None:
                    vol = random.gauss(mu=rin.vol_mu, sigma=rin.vol_sigma)
                    tmp_struc.scale_lattice(volume=vol)
                # -- save dummy coords
                dum_species = tmp_struc.species
                dum_coords = tmp_struc.cart_coords
                if rin.algo == 'EA':
                    tmp_mol_indx = []
                    tmp_id = []
                    tmp_id_cnt = 0
                # -- remove dummy
                tmp_struc.remove_sites(range(0, len(dum_species)))
                tmp_struc_ori = tmp_struc.copy()
                rot_success = False
                # -- rotate molecules
                for nrel in range(rin.nrot):
                    tmp_struc = tmp_struc_ori.copy()
                    for (dum_specie, dum_coord) in zip(dum_species, dum_coords):
                        mol_index = atype_dummy.index(dum_type[dum_specie])
                        mol = self.mol_data[mol_index]
                        # rotation option
                        if rin.rot_mol is None:
                            rot_mol_coord = mol.cart_coords
                        if rin.rot_mol == 'random':
                            angle = 2 * np.pi * np.random.rand(3)
                            R = rot_mat(angle)
                            rot_mol_coord = np.matmul(mol.cart_coords, R)
                        if rin.rot_mol == 'random_mol':
                            # each mol_data
                            mol_angles = []    # [angles of mol 1, angles of mol 2, ...]
                            for i in range(len(self.mol_data)):
                                mol_angles.append(2 * np.pi * np.random.rand(3))
                            angle = mol_angles[mol_index]
                            R = rot_mat(angle)
                            rot_mol_coord = np.matmul(mol.cart_coords, R)
                        if rin.rot_mol == 'random_wyckoff':
                            # each Wyckoff
                            dum_angles = {}    # e.g.
                                               # {DummySpecie X00+: array([ , , ]),
                                               #  DummySpecie X10+: array([ , , ]),
                                               #  DummySpecie X20+: array([ , , ]),
                                               #  DummySpecie X30+: array([ , , ])}
                            angles = 2 * np.pi * np.random.rand(len(dums), 3)
                            for (dum, angle) in zip(dums, angles):
                                dum_angles[dum] = angle
                            angle = dum_angles[dum_specie]
                            R = rot_mat(angle)
                            rot_mol_coord = np.matmul(mol.cart_coords, R)
                        # rotate coord
                        coord = rot_mol_coord + dum_coord
                        # append mol
                        for i, ms in enumerate(mol.species):
                            tmp_struc.append(ms, coord[i], coords_are_cartesian=True)
                            if rin.algo == 'EA':
                                tmp_id.append(tmp_id_cnt)
                                tmp_mol_indx.append(mol_index)
                        if rin.algo == 'EA':
                            tmp_id_cnt += 1
                    # -- check nat
                    if not self._check_nat(tmp_struc):
                        # pyxtal returns conventional cell,
                        # too many atoms if centering
                        tmp_struc = tmp_struc.get_primitive_structure()
                        # recheck nat
                        if not self._check_nat(tmp_struc):    # failure
                            if rin.rot_mol is None:
                                break    # go back to the while loop
                            continue
                    # -- sort, necessary in molecular crystal
                    tmp_struc = sort_by_atype(tmp_struc, rin.atype)
                    # -- check minimum distance
                    if self.mindist is not None:
                        success, mindist_ij, dist = check_distance(tmp_struc,
                                                                   rin.atype,
                                                                   self.mindist)
                        if not success:
                            type0 = rin.atype[mindist_ij[0]]
                            type1 = rin.atype[mindist_ij[1]]
                            logger.warning(f'mindist: {type0} - {type1}, {dist}. retry.')
                            if rin.rot_mol is None:
                                break    # go back to the while loop
                            continue    # failure
                    # -- check actual space group (success)
                    try:
                        spg_sym, spg_num = tmp_struc.get_space_group_info(
                            symprec=rin.symprec)
                    except TypeError:
                        spg_num = 0
                        spg_sym = None
                    rot_success = True
                    break
                # -- reach maximum times to rotate (failure)
                if not rot_success:
                    continue    # go back to the while loop
                # -- register the structure in pymatgen format
                cid = len(self.init_struc_data) + id_offset
                self.init_struc_data[cid] = tmp_struc
                if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
                    # calculate interatomic distance of mol_data
                    mol_dists = []
                    for j, mol in enumerate(self.mol_data):
                        tmp_dists = []
                        for n, m in enumerate(mol):
                            tmp_dists.append(mol.get_distance(0, n))
                        mol_dists.append(tmp_dists)
                    self.struc_mol_id.update({cid: [tmp_mol_indx, tmp_id, mol_dists]})
                logger.info(f'Structure ID {cid:>6} was generated.'
                      f' Space group: {spg:>3} --> {spg_num:>3} {spg_sym}')
                # -- save init_POSCARS
                if init_pos_path is not None:
                    out_poscar(tmp_struc, cid, init_pos_path)

    def _set_tol_mat(self, atype, mindist):
        tolmat = Tol_matrix()
        for i, itype in enumerate(atype):
            for j, jtype in enumerate(atype):
                if i <= j:
                    tolmat.set_tol(itype, jtype, mindist[i][j])
        return tolmat

    def _check_nat(self, struc):
        # ---------- count number of atoms in each element for check
        species_list = [a.species_string for a in struc]
        for i in range(len(rin.atype)):
            if species_list.count(rin.atype[i]) != rin.nat[i]:
                return False    # failure
        return True

    def _mp_mc(self, tolmat, spg, nmol, rand_vol, q, algo):
        '''
        multiprocess part
        here cannot use rin.xxx and logging
        '''
        try:
            np.random.seed(random.randint(0, 100000000))
            tmp_crystal = pyxtal(molecular=True)
            f = StringIO()
            with redirect_stdout(f):
                with redirect_stderr(f):
                    tmp_crystal.from_random(dim=3, group=spg,
                                            species=self.mol_data, numIons=nmol,
                                            factor=rand_vol, conventional=False, tm=tolmat)
            s = f.getvalue().rstrip()    # to delete \n
            if s:
                logger.warning(s)
            if algo == 'EA':
                tmp_struc = tmp_crystal.to_pymatgen(resort=False)
                tmp_lattice = tmp_struc.lattice
                dums = []        # dummy atoms
                dum_pos = []     # internal position of dummy
                in_dists = []
                mol_dists = []
                for i, site in enumerate(tmp_crystal.mol_sites):
                    for j, mol in enumerate(self.mol_data):
                        if mol.species == site.mol.species:
                            dum = DummySpecie("X{}{}".format(j, i))
                            dums.append(dum)
                            dum_pos.append(site.position)
                            # -- calculate atom distance by pyxtal
                            in_dists.append(np.linalg.norm(site.position.dot(tmp_crystal.lattice.get_matrix()) -
                                                           site.get_coords_and_species(absolute=True)[0]
                                                           [0:len(site.mol.species)], axis=1))
                # -- calculate atom distance from self.mol_data
                for j, mol in enumerate(self.mol_data):
                    tmp_dists = []
                    for n, m in enumerate(mol):
                        tmp_dists.append(mol.get_distance(0, n))
                    mol_dists.append(tmp_dists)
                dum_struc = Structure.from_spacegroup(spg, tmp_lattice, dums, dum_pos)
                for ds, dc in zip(dum_struc.species, dum_struc.cart_coords):
                    tmp_struc.append(ds, dc, coords_are_cartesian=True)
            # ---------- queue
            if tmp_crystal.valid:
                if algo == 'EA':
                    q.put([tmp_struc, dums, in_dists, mol_dists])
                else:
                    q.put(tmp_crystal.to_pymatgen(resort=False))
                q.put(tmp_crystal.valid)
            else:
                q.put(None)
                q.put(tmp_crystal.valid)
        except Exception as e:
            q.put('error')
            q.put(e)



'''
Utility for structures
'''

from logging import getLogger
import os

import numpy as np

from pymatgen.core import Structure
from pymatgen.io.cif import CifWriter
from pyxtal.tolerance import Tol_matrix

from ..IO import read_input as rin


logger = getLogger('cryspy')

def set_mindist(mindist_in, factor, dummy=False, mpi_rank=0):
    # ---------- dummy atom in mol_bs
    if dummy:
        atype = get_atype_dummy()
    else:
        atype = rin.atype

    # ---------- mindist
    if mindist_in is None:
        # ------ Tol matrix
        if rin.struc_mode in ['crystal']:
            tolmat = Tol_matrix(prototype='atomic', factor=factor)
        elif rin.struc_mode in ['mol', 'mol_bs']:
            tolmat = Tol_matrix(prototype='molecular', factor=factor)
        else:
            logger.error('struc_mode is wrong')
            raise SystemExit(1)
        # ------ set mindist
        mindist = []
        for i, itype in enumerate(atype):
            tmp = []
            for j, jtype in enumerate(atype):
                tmp.append(tolmat.get_tol(itype, jtype))
            mindist.append(tmp)
    else:
        tmp_array = factor * np.array(mindist_in)
        mindist = tmp_array.tolist()

    # ---------- log
    if mpi_rank == 0:
        for i, itype in enumerate(atype):
            for j, jtype in enumerate(atype):
                if i <= j:
                    if dummy:
                        logger.info(f'{rin.mol_file[i]} - {rin.mol_file[j]}: {mindist[i][j]}')
                    else:
                        logger.info(f'{itype} - {jtype}: {mindist[i][j]}')
        if rin.struc_mode == 'mol':
            logger.info('When struc_mode is mol (only random structure, not EA part),\n'
            '- tolerance between monoatomic molecules is multiplied by 0.8 inside pyxtal (not printed above)\n'
            '- H-N, H-O, or H-F tolerance is multiplied by 0.9 inside pyxtal (not printed above)')

    # ---------- return
    return mindist


def get_atype_dummy():
    noble_gas = ['Rn', 'Xe', 'Kr', 'Ar', 'Ne', 'He']
    if len(rin.nmol) > len(noble_gas):
        logger.error('len(nmol) > len(noble_gas)')
        raise SystemExit(1)
    atype = noble_gas[:len(rin.nmol)]
    return atype


def out_poscar(struc, cid, fpath):
    # ---------- poscar format
    pos = struc.to(fmt='poscar')
    pos = pos.split('\n')
    blank_indx = pos.index('')    # cut unnecessary parts
    pos = pos[:blank_indx]
    pos[0] = 'ID_{}'.format(cid)    # replace with ID
    lines = [line+'\n' for line in pos]

    # ---------- append POSCAR
    with open(fpath, 'a') as f:
        for line in lines:
            f.write(line)


def out_cif(struc, cid, tmp_path, fpath, symprec=0.1):
    # ---------- opt_CIFS
    cif = CifWriter(struc, symprec=symprec)
    cif.write_file(tmp_path+'tmp.cif')

    # ---------- correct title for VESTA
    #                (need to delete '_chemical_formula_sum'. i don't know why)
    with open(tmp_path+'tmp.cif', 'r') as fcif:
        ciflines = fcif.readlines()
    ciflines[1] = 'data_ID_{}\n'.format(cid)
    if ciflines[11][:21] == '_chemical_formula_sum':
        ciflines.pop(11)
    else:
        logger.error('ciflines[11] is not _chemical_formula_sum,'
                         ' have to fix bag')
        raise SystemExit(1)

    # ---------- cif --> opt_cifs
    with open(fpath, 'a') as foptcif:
        for line in ciflines:
            foptcif.write(line)

    # ---------- clean tmp.cif
    os.remove(tmp_path+'tmp.cif')


def frac_coord_zero_one(struc_in):
    '''
    fractional coordinates: 0.0 <= x,y,z < 1.0
    e.g. [0.0, -0.25, 0.7] --> [0.0, 0.75, 0.7]

    # ---------- args
    struc_in: structure data in pymatgen format
    '''
    struc = struc_in.copy()
    for i in range(struc.num_sites):
        struc[i] = struc[i].to_unit_cell()
    return struc


def origin_shift(struc_in):
    '''
    Randomly shift the origin of struc_in

    # ---------- args
    struc_in: structure data in pymatgen format

    # ---------- return
    origin shifted structure (not change original struc_in)
    '''
    struc = struc_in.copy()
    coords_trans = struc.frac_coords + np.random.rand(3)
    struc_shift = Structure(struc.lattice, struc.species, coords_trans)
    struc_shift = frac_coord_zero_one(struc_shift)
    return struc_shift


def sort_by_atype(struc, atype):
    '''
    return a structre sorted by atype order as a new structure
    '''
    return struc.get_sorted_structure(
        key=lambda x: atype.index(x.species_string))


def check_distance(struc, atype, mindist, check_all=False):
    '''
    # ---------- args
    struc: structure data in pymatgen format
    atype (list): e.g. ['Li', 'Co, 'O']
    mindist (2d list) : e.g. [[2.0, 2.0, 1.2],
                              [2.0, 2.0, 1.2],
                              [1.2, 1.2, 1.5]]
    check_all (bool) : if True, check all atom pairs, return dist_list.
                       if False, stop when (dist < mindist) is found,
                                 return True or False (see below)
    # ---------- return
    (check_all=False) True, None, None: nothing smaller than mindist
    (check_all=False) False, (i, j), dist: something smaller than mindst
                                     here, (i, j) means mindist(i, j)
    (check_all=True) dist_list: if dist_list is vacant,
                                    nothing smaller than mindist
    '''

    # ---------- initialize
    if check_all:
        dist_list = []    # [(i, j, dist), (i, j, dist), ...]

    # ---------- in case there is only one atom
    if struc.num_sites == 1:
        dist = min(struc.lattice.abc)
        if dist < mindist[0][0]:
            if check_all:
                dist_list.append((0, 0, dist))
                return dist_list
            return False, (0, 0), dist
        return True, None, None

    # ---------- normal case
    for i in range(struc.num_sites):
        for j in range(i):
            dist = struc.get_distance(j, i)
            i_type = atype.index(struc[i].species_string)
            j_type = atype.index(struc[j].species_string)
            if dist < mindist[i_type][j_type]:
                if check_all:
                    dist_list.append((j, i, dist))
                else:
                    return False, (i_type, j_type), dist

    # ---------- return
    if check_all:
        if dist_list:
            dist_list.sort()    # sort
            return dist_list
        else:
            return dist_list    # dist_list is vacant list
    return True, None, None


def scale_cell_mol(struc, mol_data, volume, dist_digit=3):
    '''
    scale cell without changing molecule size
    for molecular structures generated by PyXtal
    '''
    # ---------- get molecules
    molecules = []
    for mol in mol_data:
        if len(mol.species) > 1:    # mol.species == 1 <-- atom
            molecules.append(mol)

    # ---------- structure of molecule
    mols_cart_coords = []
    mols_frac_Gs = []
    for mol in molecules:
        # ------ start atom for the molecule
        mol_pos = []
        for i, element in enumerate(struc.species):
            # -- check len
            if len(struc.species) < mol.num_sites + i:
                break
            # -- check number of atoms
            count = 0
            for j in range(mol.num_sites):
                if struc.species[i+j] == mol.species[j]:
                    count += 1
            if count == mol.num_sites:
                mol_pos.append(i)
        # ------ calculate the center of gravity
        in_dists = np.linalg.norm(mol.cart_coords - mol.cart_coords[0], axis=1)
        frac_Gs = []
        mol_cart_coords = []
        # -- take periodic boundary into account
        #    here, check distance from the reference atom (mol.cart_coords[0])
        #    to find a correct site for calculating the center of gravity
        for st in mol_pos:
            frac_mol_coords = struc.frac_coords[st:st + mol.num_sites]
            cart_mol_basis = struc.lattice.get_cartesian_coords(frac_mol_coords[0])
            fix_frac_coords = []
            for frac, dist in zip(frac_mol_coords, in_dists):
                all_pattern = [[frac[0]], [frac[1]], [frac[2]]]
                basis_dist = frac - frac_mol_coords[0]
                for axis in range(3):
                    if basis_dist[axis] > 0:
                        all_pattern[axis].append(frac[axis]-1)
                    elif basis_dist[axis] < 0:
                        all_pattern[axis].append(frac[axis]+1)
                # check all patterns
                break_loop = False
                for a_coord in all_pattern[0]:
                    for b_coord in all_pattern[1]:
                        for c_coord in all_pattern[2]:
                            tmp_frac_coord = [a_coord, b_coord, c_coord]
                            tmp_cart_coord = struc.lattice.get_cartesian_coords(tmp_frac_coord)
                            tmp_cart_dist = np.linalg.norm(tmp_cart_coord - cart_mol_basis)
                            if round(dist, dist_digit) == round(tmp_cart_dist, dist_digit):
                                break_loop = True
                                fix_frac_coords.append(tmp_frac_coord)
                                break
                        if break_loop:
                            break
                    if break_loop:
                        break
            if not len(fix_frac_coords) == mol.num_sites:
                return False
            fix_frac_coords = np.array(fix_frac_coords)
            # -- get the center of gravity
            frac_G = fix_frac_coords.sum(axis=0)/mol.num_sites
            frac_Gs.append(frac_G)
            # -- save structure of mol
            fix_cart_coords = struc.lattice.get_cartesian_coords(fix_frac_coords)
            cart_G = fix_cart_coords.sum(axis=0)/mol.num_sites
            mol_cart_coord = fix_cart_coords - cart_G
            mol_cart_coords.append(mol_cart_coord)
        mols_frac_Gs.append(frac_Gs)
        mols_cart_coords.append(mol_cart_coords)
        # ------ remove the molecule from sturc
        for st in reversed(mol_pos):
            struc.remove_sites(range(st, st + mol.num_sites))

    # ---------- scale lattice
    struc.scale_lattice(volume)

    # ---------- put the molecule of original size
    for mol_index in range(len((molecules))):
        molecule = molecules[mol_index]
        for frac_G, cart_mol in zip(mols_frac_Gs[mol_index],
                                    mols_cart_coords[mol_index]):
            cart_G = struc.lattice.get_cartesian_coords(frac_G)
            cart_mol_coord = cart_mol + cart_G
            for spe, spe_coord in zip(molecule.species, cart_mol_coord):
                struc.append(spe, spe_coord, coords_are_cartesian=True)

    # ---------- return new struc
    return struc


def rot_mat(angles, seq='zyx', degree=False):
    '''
    Calculate rotation matrix

    # ---------- input
    angles: in radian (degree=False)
    seq:
    degree:
    '''
    if not len(seq) == len(angles):
        logger.error('not len(seq) == len(angles)')
        raise SystemExit(1)

    if degree:
        angles = np.deg2rad(angles)

    R = np.eye(3)
    angles = angles[::-1]
    seq = seq[::-1]
    for axis, angle in zip(seq, angles):
        if 'x' == axis:
            tmp_R = np.array([[1, 0, 0],
                              [0, np.cos(angle), -np.sin(angle)],
                              [0, np.sin(angle), np.cos(angle)]])
        elif 'y' == axis:
            tmp_R = np.array([[np.cos(angle), 0, np.sin(angle)],
                              [0, 1, 0],
                              [-np.sin(angle), 0, np.cos(angle)]])
        elif 'z' == axis:
            tmp_R = np.array([[np.cos(angle), -np.sin(angle), 0],
                              [np.sin(angle), np.cos(angle), 0],
                              [0, 0, 1]])
        else:
            tmp_R = np.eye(3)
        R = np.matmul(R, tmp_R)

    return R


def find_site(struc, mol_id, group_id, true_dists):
    '''
    find out atom on the unit cell, because don't refer atom in neighbor cell
    '''
    # ---------- init
    fix_frac_coords = []
    fix_group_id = []
    fix_mol_id = []
    grouped_struc_mol_id = []
    id_cnt = 0
    grouped_struc = []
    # ----------  separate molecules per group_id
    for j in range(len(struc)):
        remove_indx = []
        for i, s in enumerate(struc):
            if id_cnt != group_id[i]:
                remove_indx.append(i)
        if len(remove_indx) != len(struc):
            tmp_struc = struc.copy()
            tmp_struc.remove_sites(remove_indx)
            grouped_struc.append(tmp_struc)
            for n in range(len(struc)):
                if group_id[n] == id_cnt:
                    grouped_struc_mol_id.append(mol_id[n])
                    break
        id_cnt += 1
    fix_species = []
    # ---------- compare atom distance to atom distance of mol_data
    for j, ts in enumerate(grouped_struc):
        for i, frac in enumerate(ts.frac_coords):
            dist = true_dists[grouped_struc_mol_id[j]][i]
            all_pattern = [[frac[0]], [frac[1]], [frac[2]]]
            basis_dist = frac - ts[0].frac_coords
            for axis in range(3):
                if basis_dist[axis] >= 0:
                    all_pattern[axis].append(frac[axis]-1)
                elif basis_dist[axis] < 0:
                    all_pattern[axis].append(frac[axis]+1)
            # ------ check all patterns
            # break_loop = False
            min_dist = 1000000
            for a_coord in all_pattern[0]:
                for b_coord in all_pattern[1]:
                    for c_coord in all_pattern[2]:
                        tmp_frac_coord = [a_coord, b_coord, c_coord]
                        tmp_cart_coord = struc.lattice.get_cartesian_coords(tmp_frac_coord)
                        base_cart = struc.lattice.get_cartesian_coords(ts.frac_coords[0])
                        tmp_cart_dist = np.linalg.norm(tmp_cart_coord - base_cart)
                        dist_rmse = np.sqrt(np.mean((dist - tmp_cart_dist)**2))
                        if dist_rmse < min_dist:
                            min_dist = dist_rmse
                            min_dist_frac_coord = tmp_frac_coord
            fix_frac_coords.append(min_dist_frac_coord)
            fix_group_id.append(j)
            for n, gid in enumerate(group_id):
                if gid == j:
                    fix_mol_id.append(grouped_struc_mol_id[j])
                    break
            fix_species.append(grouped_struc[j][i].specie)
    return np.array(fix_frac_coords), fix_group_id, fix_species, fix_mol_id


def cal_g(struc, mol_id, group_id, true_dists):
    # ----------  calculate molecule center
    mol_g_A = []
    frac_coords, fix_group_id, _, _ = find_site(struc, mol_id, group_id, true_dists)
    for cnt in range(len(struc)):
        acnt = 0
        cal_g = 0
        for i, fc in enumerate(frac_coords):
            if cnt == fix_group_id[i]:
                acnt += 1
                cal_g += fc
        if acnt == 0:
            break
        mol_g_A.append(cal_g/acnt)
    return mol_g_A


def sort_by_atype_mol(struc, atype, mol_id, group_id):
    # ---------- work the same as sort_by_atype, and sort mol_id and group_id
    copy_struc = struc.copy()
    sorted_struc = sort_by_atype(struc, atype)
    sorted_mol_id = []
    sorted_group_id = []
    for i in range(len(sorted_struc)):
        for j in range(len(copy_struc)):
            if np.allclose(sorted_struc.frac_coords[i], copy_struc.frac_coords[j]):
                sorted_mol_id.append(mol_id[j])
                sorted_group_id.append(group_id[j])
    return sorted_struc, sorted_mol_id, sorted_group_id



'''
Modify Policy class in combo3 for CrySPY

 This code partly includes program codes in COMBO
 (https://github.com/tsudalab/combo)
 which is distributed under the MIT License.
'''

from logging import getLogger

import combo
import numpy as np


logger = getLogger('cryspy')

# ---------- inheritance
class Policy_cryspy(combo.search.discrete.policy):

    def specified_search(self, specified_actions, max_num_probes,
                         num_search_each_probe=1,
                         simulator=None, is_disp=True):
        N = int(num_search_each_probe)
        if int(max_num_probes) * N > len(self.actions):
            logger.error('max_num_probes * num_search_each_probe must'
                        ' be smaller than the length of candidates')
            raise SystemExit(1)
        for n in range(0, max_num_probes):
            action = self.get_specified_action(N, specified_actions)
            return action

    def get_specified_action(self, N, specified_actions):
        action = specified_actions
        self.actions = self.delete_actions(action)
        return action

    def bayes_search_cryspy(self, training=None, max_num_probes=None,
                            num_search_each_probe=1,
                            predictor=None, is_disp=True,
                            simulator=None, score='TS', interval=0,
                            num_rand_basis=0):
        if max_num_probes is None:
            max_num_probes = 1
#            simulator = None
        is_rand_expans = False if num_rand_basis == 0 else True
        self.training = self._set_training(training)
        if predictor is None:
            self.predictor = self._init_predictor(is_rand_expans)
        else:
            self.predictor = predictor
        N = int(num_search_each_probe)
        for n in range(max_num_probes):
            if combo.search.utility.is_learning(n, interval):
                self.predictor.fit(self.training, num_rand_basis)
                self.test.Z = self.predictor.get_basis(self.test.X)
                self.training.Z = self.predictor.get_basis(self.training.X)
                self.predictor.prepare(self.training)
            else:
                try:
                    self.predictor.update(self.training, self.new_data)
                except:
                    self.predictor.prepare(self.training)
            if num_search_each_probe != 1:
                combo.search.utility.show_start_message_multi_search(
                    self.history.num_runs, score)
            K = self.config.search.multi_probe_num_sampling
            alpha = self.config.search.alpha
            # ---------- start cryspy
            (action, cryspy_mean,
             cryspy_var, cryspy_score) = self.get_actions_cryspy(score, N, K, alpha)
            return action, cryspy_mean, cryspy_var, cryspy_score
            # ---------- end cryspy
#            if simulator is None:
#                return action
#            t, X = call_simulator(simulator, action)
#            self.write(action, t, X)
#            if is_disp:
#                combo.search.utility.show_search_results(self.history, N)
#        return copy.deepcopy(self.history)

    def get_actions_cryspy(self, mode, N, K, alpha):
#        f = self.get_score_cryspy(mode, self.predictor, self.training, alpha)
        # ---------- start cryspy
        f, cryspy_mean, cryspy_var = self.get_score_cryspy(
            mode, self.predictor, self.training, alpha)
        cryspy_score = f
        # ---------- end cryspy
        temp = np.argmax(f)
        action = self.actions[temp]
        self.actions = self.delete_actions(temp)
        chosed_actions = np.zeros(N, dtype=int)
        chosed_actions[0] = action
        for n in range(1, N):
            f = self.get_marginal_score(mode, chosed_actions[0:n], K, alpha)
            temp = np.argmax(np.mean(f, 0))
            chosed_actions[n] = self.actions[temp]
            self.actions = self.delete_actions(temp)
        return chosed_actions, cryspy_mean, cryspy_var, cryspy_score

    def get_score_cryspy(self, mode, predictor=None, training=None, alpha=1):
        self._set_training(training)
        self._set_predictor(predictor)
        actions = self.actions
        test = self.test.get_subset(actions)
        if mode == 'EI':
            f = combo.search.score.EI(predictor, training, test)
        elif mode == 'PI':
            f = combo.search.score.PI(predictor, training, test)
        elif mode == 'TS':
            f = combo.search.score.TS(predictor, training, test, alpha)
        else:
            logger.error('mode must be EI, PI or TS.')
            raise SystemExit(1)
        # ---------- start cryspy
        cryspy_mean = predictor.get_post_fmean(training, test)
        cryspy_var = predictor.get_post_fcov(training, test)
        return f, cryspy_mean, cryspy_var
        # ---------- end cryspy
#        return f



'''
Restart CrySPY
'''

from datetime import datetime
from logging import getLogger
import os
import sys

from .gen_init_struc import gen_init_struc
from ..IO import io_stat, pkl_data
from ..IO import read_input as rin
from ..util.utility import get_version, backup_cryspy

# ---------- import later
#from ..RS import rs_restart
#from ..BO import bo_restart
#from ..LAQA import laqa_restart
#from ..EA import ea_append
#from ..RS.gen_struc_RS.gen_pyxtal import Rnd_struc_gen_pyxtal
#from ..RS.gen_struc_RS.random_generation import Rnd_struc_gen


logger = getLogger('cryspy')

def restart(comm, mpi_rank, mpi_size):
    if mpi_rank == 0:
        logger.info('\n\n\nRestart CrySPY ' + get_version() + '\n\n')
        # ---------- read stat
        stat = io_stat.stat_read()
    else:
        stat = None

    # ########## MPI start
    if mpi_size > 1:
        comm.barrier()
    # ---------- read input and check the change
    try:
        rin.readin()
    except Exception as e:
        if mpi_rank == 0:
            logger.error(e.args[0])
        raise SystemExit(1)
    if mpi_rank == 0:
        try:
            rin.diffinstat(stat)
        except Exception as e:
            if mpi_size > 1:
                comm.Abort(1)
            logger.error(e.args[0])
            raise SystemExit(1)

    # ------ load init_struc_data for appending structures
    # In EA, one can not change tot_struc, so struc_mol_id need not be considered here
    init_struc_data = pkl_data.load_init_struc()

    # ---------- append structures
    if len(init_struc_data) < rin.tot_struc:
        # ------ backup
        if mpi_rank == 0:
            backup_cryspy()
        # ------ barrier for MPI
        if mpi_size > 1:
            comm.barrier()
        # ------ append struc.
        if mpi_rank == 0:
            prev_nstruc = len(init_struc_data)
        init_struc_data = _append_struc(init_struc_data, comm, mpi_rank, mpi_size)
        # ------ post append
        if mpi_rank == 0:
            # -- RS
            if rin.algo == 'RS':
                from ..RS import rs_restart
                rs_restart.restart(stat, prev_nstruc)
            # -- BO
            if rin.algo == 'BO':
                from ..BO import bo_restart
                bo_restart.restart(init_struc_data, prev_nstruc)
            # -- LAQA
            if rin.algo == 'LAQA':
                from ..LAQA import laqa_restart
                laqa_restart.restart(stat, prev_nstruc)
            os.remove('lock_cryspy')
        raise SystemExit()
    elif rin.tot_struc < len(init_struc_data):
        logger.error('tot_struc < len(init_struc_data)')
        raise SystemExit(1)

    # ---------- append structures by EA (option)
    # not support MPI
    if rin.append_struc_ea:
        if mpi_rank == 0:
            # ------ backup
            backup_cryspy()
    #
    # struc_mol_id has not developed yet here
    #if rin.struc_mode in ['mol', 'mol_bs']:
    #    struc_mol_id = pkl_data.load_struc_mol_id()
    #
            from ..EA import ea_append
            prev_nstruc = len(init_struc_data)
            init_struc_data = ea_append.append_struc(stat, init_struc_data)
            # ------ RS
            if rin.algo == 'RS':
                from ..RS import rs_restart
                rs_restart.restart(stat, prev_nstruc)
            # ------ BO
            if rin.algo == 'BO':
                from ..BO import bo_restart
                bo_restart.restart(init_struc_data, prev_nstruc)
            # ------ LAQA
            if rin.algo == 'LAQA':
                from ..LAQA import laqa_restart
                laqa_restart.restart(stat, prev_nstruc)
            os.remove('lock_cryspy')
        raise SystemExit()

    # ---------- return
    return stat, init_struc_data


def _append_struc(init_struc_data, comm, mpi_rank, mpi_size):
    # ---------- append initial structures
    if mpi_rank == 0:
        logger.info('# ---------- Append structures')
    if mpi_size > 1:
        comm.barrier()
    # ------ time
    if mpi_rank == 0:
        time_start = datetime.today()

    # ---------- gen_init_struc()
    # only init_struc_data in rank0 is important
    init_struc_data = gen_init_struc(init_struc_data, None,
                                        comm, mpi_rank, mpi_size)

    if mpi_rank == 0:
        # ---------- save
        pkl_data.save_init_struc(init_struc_data)
        # ---------- time
        time_end = datetime.today()
        etime = time_end - time_start
        logger.info(f'Elapsed time for structure generation: {etime}')

    # ---------- return
    # only init_struc_data in rank0 is important
    return init_struc_data


'''
Initialize random search
'''

from ..IO import io_stat, pkl_data
from ..IO import read_input as rin


def initialize(stat):
    # ---------- initialize
    id_queueing = [i for i in range(rin.tot_struc)]
    id_running = []

    # ---------- status
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- save
    rs_id_data = (id_queueing, id_running)
    pkl_data.save_rs_id(rs_id_data)



'''
Generational change in evolutionary algorithm
'''

from logging import getLogger

import pandas as pd

from .ea_child import child_gen
from .gen_struc_EA.select_parents import Select_parents
from ..IO import out_results
from ..IO import change_input, io_stat, pkl_data
from ..IO import read_input as rin


logger = getLogger('cryspy')

def next_gen(stat, init_struc_data, struc_mol_id, opt_struc_data, rslt_data, ea_id_data):
    # ---------- ea_id_data
    gen, id_queueing, id_running = ea_id_data

    # ---------- log
    logger.info('# ---------- Evolutionary algorithm')
    logger.info(f'# ------ Generation {gen + 1}')

    # ---------- current generation
    c_rslt = rslt_data[rslt_data['Gen'] == gen]
    c_fitness = c_rslt['E_eV_atom'].to_dict()    # {ID: energy, ...}

    # ---------- load ea_data, ea_data is used only in this module
    elite_struc, elite_fitness, ea_info, ea_origin = pkl_data.load_ea_data()

    # ---------- instantiate Seclect_parents class
    logger.info('# -- select parents')
    sp = Select_parents(opt_struc_data, c_fitness, elite_struc, elite_fitness, rin.n_fittest)
    if rin.slct_func == 'TNM':
        sp.set_tournament()
    else:
        sp.set_roulette()

    # ---------- generate offspring by EA
    logger.info('# -- Generate structures')
    if rin.struc_mode not in ['mol', 'mol_bs']:
        _, eagen = child_gen(sp, init_struc_data, None)
    else:
        _, eagen, _ = child_gen(sp, init_struc_data, struc_mol_id)

    # ---------- select elite
    if rin.n_elite > 0:
        logger.info('# -- select elites')
        # ------ init
        all_fitness = rslt_data['E_eV_atom'].to_dict()    # {ID: energy, ..,}
        elite_struc = {}
        elite_fitness = {}
        # ------ Select_parents class also works for selecting elite structures
        se = Select_parents(opt_struc_data, all_fitness, None, None, rin.n_elite)
        for cid in se.ranking_dedupe:
            logger.info(f'Structure ID {cid:>6} keeps as the elite')
            elite_struc[cid] = opt_struc_data[cid]
            elite_fitness[cid] = all_fitness[cid]
    else:
        elite_struc = None
        elite_fitness = None

    # ---------- new generation
    gen += 1

    # ---------- id_queueing
    id_queueing = [i for i in range(rin.tot_struc, rin.tot_struc + rin.n_pop)]

    # ---------- ea_info
    tmp_info = pd.DataFrame(data=[[gen, rin.n_pop, rin.n_crsov, rin.n_perm,
                                   rin.n_strain, rin.n_rand, rin.n_elite,
                                   rin.crs_lat, rin.slct_func]],
                            columns=ea_info.columns)
    ea_info = pd.concat([ea_info, tmp_info], axis=0, ignore_index=True)
    # ------ out ea_info
    out_results.out_ea_info(ea_info)

    # ---------- ea_origin
    # ------ EA operation part
    for cid in range(rin.tot_struc, rin.tot_struc + rin.n_pop - rin.n_rand):
        tmp_origin = pd.DataFrame([[gen, cid, eagen.operation[cid],
                                    eagen.parents[cid]]], columns=ea_origin.columns)
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ random part
    for cid in range(rin.tot_struc + rin.n_pop - rin.n_rand,
                     rin.tot_struc + rin.n_pop):
        tmp_origin = pd.DataFrame([[gen, cid, 'random', None]],
                                  columns=ea_origin.columns)
        ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ elite part
    if rin.n_elite > 0:
        for cid in se.ranking_dedupe:
            tmp_origin = pd.DataFrame([[gen, cid, 'elite', 'elite']],
                                      columns=ea_origin.columns)
            ea_origin = pd.concat([ea_origin, tmp_origin], axis=0, ignore_index=True)
    # ------ out ea_origin
    out_results.out_ea_origin(ea_origin)

    # ---------- save ea_id_data
    ea_id_data = (gen, id_queueing, id_running)
    pkl_data.save_ea_id(ea_id_data)

    # ---------- save ea_data
    ea_data = (elite_struc, elite_fitness, ea_info, ea_origin)
    pkl_data.save_ea_data(ea_data)

    # ---------- change the value of tot_struc
    config = change_input.config_read()
    change_input.change_basic(config, 'tot_struc', rin.tot_struc + rin.n_pop)
    change_input.write_config(config)
    logger.info('# -- changed cryspy.in')
    logger.info('Changed the value of tot_struc in cryspy.in'
          f' from {rin.tot_struc} to {rin.tot_struc + rin.n_pop}')

    # ---------- status
    io_stat.set_input_common(stat, 'basic', 'tot_struc', rin.tot_struc + rin.n_pop)
    io_stat.set_common(stat, 'generation', gen)
    io_stat.set_id(stat, 'id_queueing', id_queueing)
    io_stat.write_stat(stat)

    # ---------- ext
    if rin.calc_code == 'ext':
        with open('ext/stat_job', 'w') as fstat:
            fstat.write('out\n')



'''
Collect results in ASE
'''

from logging import getLogger

import numpy as np
from pymatgen.core import Structure

from ...IO import read_input as rin


logger = getLogger('cryspy')

def collect_ase(current_id, work_path):
    # ---------- etc
    magmom = np.nan
    check_opt = 'no_file'    # always no_file in ASE for now
    # ---------- collect energy
    try:
        with open(work_path+'log.tote') as f:
            lines = f.readlines()
        energy = float(lines[-1].split()[0])    # in eV/cell
        energy = energy/float(rin.natot)    # eV/cell --> eV/atom
    except Exception as e:
        energy = np.nan    # error
        logger.warning(e.args[0] + f':    Structure ID {current_id}, could not obtain energy from log.tote')
    # ---------- collect CONTCAR
    try:
        opt_struc = Structure.from_file(work_path+'CONTCAR')
    except Exception:
        opt_struc = None
    # ---------- check
    if np.isnan(energy):
        opt_struc = None
    if opt_struc is None:
        energy = np.nan
        magmom = np.nan
    # ---------- return
    return opt_struc, energy, magmom, check_opt


'''
I/O for cryspy.stat
'''
import configparser

from . import read_input as rin


def stat_init():
    stat = configparser.ConfigParser()
    stat.add_section('basic')
    stat.add_section('structure')
    # ---------- algo
    if rin.algo == 'BO':
        stat.add_section('BO')
    if rin.algo == 'LAQA':
        stat.add_section('LAQA')
    if rin.algo == 'EA':
        stat.add_section('EA')
    # ---------- calc_code
    if rin.calc_code == 'VASP':
        stat.add_section('VASP')
    if rin.calc_code == 'QE':
        stat.add_section('QE')
    if rin.calc_code == 'soiap':
        stat.add_section('soiap')
    if rin.calc_code == 'LAMMPS':
        stat.add_section('LAMMPS')
    if rin.calc_code == 'OMX':
        stat.add_section('OMX')
    if rin.calc_code == 'ASE':
        stat.add_section('ASE')
    # ----------
    stat.add_section('option')
    stat.add_section('status')
    return stat


def stat_read():
    stat = configparser.ConfigParser()
    stat.read('cryspy.stat')
    return stat


def write_stat(stat):
    with open('cryspy.stat', 'w') as f:
        stat.write(f)


# ---------- input section
def set_input_common(stat, sec, var_str, var):
    stat.set(sec, var_str, f'{var}')


# ---------- status section
def set_common(stat, var_str, var):
    stat.set('status', var_str, f'{var}')


def set_id(stat, var_str, var_list):
    if len(var_list) > 30:
        vl = ' '.join(str(a) for a in var_list[:5])
        stat.set('status', var_str, f'{vl} ... total {len(var_list)} IDs')
    else:
        vl = ' '.join(str(a) for a in var_list)
        stat.set('status', var_str, f'{vl}')


def set_stage(stat, current_id, current_stage):
    stat.set('status', f'ID {current_id:>6}', f'Stage {current_stage}')


def clean_id(stat, current_id):
    stat.remove_option('status', f'ID {current_id:>6}')



from logging import getLogger
import os
import sys

from ..IO import read_input as rin
from ..util.utility import check_fwpath
from ..util.struc_util import set_mindist, out_poscar


logger = getLogger('cryspy')

def gen_init_struc(init_struc_data, struc_mol_id, comm, mpi_rank, mpi_size):
    # ---------- mindist
    if mpi_rank == 0:
        logger.info('# ------ mindist')
    if mpi_size > 1:
        comm.barrier()
    mindist = set_mindist(rin.mindist, rin.mindist_factor, False, mpi_rank)

    # ---------- nstruc, offset for MPI
    nstruc_list, offset_list = _divide_task(rin.tot_struc, mpi_size, len(init_struc_data))

    # ---------- log
    if mpi_rank == 0:
        logger.info('# ------ generate structures')

    # ---------- pyxtal
    if not (rin.spgnum == 0 or rin.use_find_wy):
        from ..RS.gen_struc_RS.gen_pyxtal import Rnd_struc_gen_pyxtal
        rsgx = Rnd_struc_gen_pyxtal(mindist=mindist)
        # ------ crystal
        if rin.struc_mode == 'crystal':
            rsgx.gen_struc(nstruc=nstruc_list[mpi_rank], id_offset=offset_list[mpi_rank])
        # ------ molecular crystal
        elif rin.struc_mode == 'mol':
            rsgx.set_mol()
            rsgx.gen_struc_mol(nstruc=nstruc_list[mpi_rank], id_offset=offset_list[mpi_rank])
        # ------ molecular crystal breaking symmetry
        elif rin.struc_mode == 'mol_bs':
            if mpi_rank == 0:
                logger.info('# -- mindist_mol_bs')
            mindist_dummy = set_mindist(rin.mindist_mol_bs, rin.mindist_mol_bs_factor,
                                        dummy=True, mpi_rank=mpi_rank)
            rsgx.set_mol()
            rsgx.gen_struc_mol_break_sym(nstruc=nstruc_list[mpi_rank],
                                            mindist_dummy=mindist_dummy,
                                            id_offset=offset_list[mpi_rank])
        # ------ init_struc_data
        if mpi_size > 1:
            # -- parallel
            # only init_struc_data in rank0 is important
            if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
                struc_dict, mol_id_dict = _gather_struc(rsgx.init_struc_data, rsgx.struc_mol_id,
                                                        comm, mpi_rank)
                init_struc_data.update(struc_dict)
                struc_mol_id.update(mol_id_dict)
            else:
                struc_dict, _ = _gather_struc(rsgx.init_struc_data, None, comm, mpi_rank)
                init_struc_data.update(struc_dict)
            # init_POSCARS
            if mpi_rank == 0:
                for cid, struc in struc_dict.items():
                    out_poscar(struc, cid, './data/init_POSCARS')
        else:
            # -- serial
            init_struc_data.update(rsgx.init_struc_data)
            if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
                struc_mol_id.update(rsgx.struc_mol_id)
            # init_POSCARS
            for cid, struc in rsgx.init_struc_data.items():
                out_poscar(struc, cid, './data/init_POSCARS')
    # ---------- w/o pyxtal
    else:
        from ..RS.gen_struc_RS.random_generation import Rnd_struc_gen
        rsg = Rnd_struc_gen(mindist=mindist)
        if rin.spgnum == 0:
            rsg.gen_wo_spg(nstruc=nstruc_list[mpi_rank], id_offset=offset_list[mpi_rank])
        else:
            # ---- findwy
            # -- check fwpath
            fwpath = check_fwpath(rin.fwpath)
            if mpi_rank == 0:
                os.makedirs('tmp_gen_struc', exist_ok=True)
            if mpi_size > 1:
                comm.barrier()
            rsg.gen_with_find_wy(nstruc=nstruc_list[mpi_rank],
                                    id_offset=offset_list[mpi_rank], init_pos_path=None,
                                    fwpath=fwpath, mpi_rank=mpi_rank)
        # ------ init_struc_data
        if mpi_size > 1:
            # -- parallel
            struc_dict, _ = _gather_struc(rsg.init_struc_data, None, comm, mpi_rank)
            init_struc_data.update(struc_dict)
            # init_POSCARS
            if mpi_rank == 0:
                for cid, struc in struc_dict.items():
                    out_poscar(struc, cid, './data/init_POSCARS')
        else:
            # -- serial
            init_struc_data.update(rsg.init_struc_data)
            # init_POSCARS
            for cid, struc in rsg.init_struc_data.items():
                out_poscar(struc, cid, './data/init_POSCARS')

    # ---------- return
    # only init_struc_data in rank0 is important
    if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
        return init_struc_data, struc_mol_id
    else:
        return init_struc_data


def _divide_task(n, size, init_n = 0):
    '''
    For example, n = 10, size = 4,
    we would like to divide task as follows:
        rank0 = [0, 1, 2] <-- 3
        rank1 = [3, 4, 5] <-- 3
        rank2 = [6, 7]    <-- 2
        rank3 = [8, 9]    <-- 2

    nstruc_list = [3, 3, 2, 2]
    offset_list = [0, 3, 6, 8]

    if size == 1:
        nstruc_list = [n]
        offset_list = [0]

    case: n = 20, size = 4, init_n = 10
        rank0 = [10, 11, 12] <-- 3
        rank1 = [13, 14, 15] <-- 3
        rank2 = [16, 17]    <-- 2
        rank3 = [18, 19]    <-- 2

    nstruc_list = [3, 3, 2, 2]
    offset_list = [10, 13, 16, 18]

    if size == 1:
        nstruc_list = [10] <-- n - init_n
        offset_list = [10] <-- init_n
    '''
    q = (n - init_n)//size
    r = (n - init_n)%size
    nstruc_list = []
    offset_list = []
    offset = init_n
    for i in range(size):
        offset_list.append(offset)
        if r > 0:
            nstruc_list.append(q + 1)
            offset += q + 1
            r = r -1
        else:
            nstruc_list.append(q)
            offset += q
    return nstruc_list, offset_list


def _gather_struc(struc_dict, mol_id_dict, comm, mpi_rank):
    '''
    only if mpi_size > 1
    '''
    # ---------- struc_data
    data = struc_dict
    data = comm.gather(data, root=0)
    if mpi_rank == 0:
        for d in data:
            struc_dict.update(d)
    else:
        assert data is None
    # ---------- for mol_id
    if rin.algo == 'EA' and rin.struc_mode in ['mol', 'mol_bs']:
        data = mol_id_dict
        data = comm.gather(data, root=0)
        if mpi_rank == 0:
            for d in data:
                mol_id_dict.update(d)
        else:
            assert data is None
    # ---------- return
    return struc_dict, mol_id_dict